# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Base and auxiliary classes for FS and project browsers"""

import os.path
import logging
from utils.globals import GlobalData
from utils.pixmapcache import getIcon
from utils.fileutils import isPythonMime, isCDMProjectMime, getFileProperties
from utils.project import getProjectFileTooltip
from .qt import (Qt, QModelIndex, QSortFilterProxyModel, QAbstractItemView,
                 QApplication, QTreeView, pyqtSignal)
from .viewitems import (DirectoryItemType, SysPathItemType, GlobalsItemType,
                        ImportsItemType, FunctionsItemType, ClassesItemType,
                        StaticAttributesItemType, InstanceAttributesItemType,
                        CodingItemType, ImportItemType, FileItemType,
                        FunctionItemType, ClassItemType, DecoratorItemType,
                        AttributeItemType, GlobalItemType, ImportWhatItemType,
                        TreeViewDirectoryItem, TreeViewFileItem,
                        TreeViewCodingItem, TreeViewGlobalsItem,
                        TreeViewGlobalItem, TreeViewImportsItem,
                        TreeViewImportItem, TreeViewWhatItem,
                        TreeViewFunctionsItem, TreeViewClassesItem)
from .itemdelegates import NoOutlineHeightDelegate
from .parsererrors import ParserErrorsDialog
from .findinfiles import FindInFilesDialog


class FilesBrowserSortFilterProxyModel(QSortFilterProxyModel):

    """Files (filesystem and project) browser sort filter proxy model.

    It allows filtering basing on top level items.
    """

    def __init__(self, isProjectFilter, parent=None):
        QSortFilterProxyModel.__init__(self, parent)
        self.__sortColumn = None    # Avoid pylint complains
        self.__sortOrder = None     # Avoid pylint complains
        self.__shouldFilter = isProjectFilter

    def sort(self, column, order):
        """Sorts the items"""
        self.__sortColumn = column
        self.__sortOrder = order
        QSortFilterProxyModel.sort(self, column, order)

    def lessThan(self, left, right):
        """Sorts the displayed items"""
        lhs = left.model() and left.model().item(left) or None
        rhs = right.model() and right.model().item(right) or None

        if lhs and rhs:
            return lhs.lessThan(rhs, self.__sortColumn, self.__sortOrder)
        return False

    def item(self, index):
        """Provides a reference to the item"""
        if not index.isValid():
            return None

        sourceIndex = self.mapToSource(index)
        return self.sourceModel().item(sourceIndex)

    def hasChildren(self, parent=QModelIndex()):
        """Checks the presence of the child items"""
        sourceIndex = self.mapToSource(parent)
        return self.sourceModel().hasChildren(sourceIndex)

    def filterAcceptsRow(self, sourceRow, sourceParent):
        """Filters rows"""
        if not self.__shouldFilter:
            return True     # Show everything

        # Filter using the loaded project filter
        if not sourceParent.isValid():
            return True

        item = sourceParent.internalPointer().child(sourceRow)
        return not GlobalData().project.shouldExclude(item.data(0))


class FilesBrowser(QTreeView):

    """Common functionality of the FS and project browsers"""

    sigFirstSelectedItem = pyqtSignal(QModelIndex)

    def __init__(self, sourceModel, isProjectFilter, parent=None):
        QTreeView.__init__(self, parent)

        self.__model = sourceModel
        self.__sortModel = FilesBrowserSortFilterProxyModel(isProjectFilter)
        self.__sortModel.setSourceModel(self.__model)
        self.setModel(self.__sortModel)

        self.activated.connect(self.openSelectedItem)
        self.expanded.connect(self._resizeColumns)
        self.collapsed.connect(self._resizeColumns)

        self.setRootIsDecorated(True)
        self.setAlternatingRowColors(True)
        self.setUniformRowHeights(True)
        self.setItemDelegate(NoOutlineHeightDelegate(4))

        header = self.header()
        header.setSortIndicator(0, Qt.AscendingOrder)
        header.setSortIndicatorShown(True)
        header.setSectionsClickable(True)

        self.setSortingEnabled(True)

        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.header().setStretchLastSection(True)
        self.layoutDisplay()

        self.__debugMode = False

    def layoutDisplay(self):
        """Performs the layout operation"""
        self._resizeColumns(QModelIndex())
        self._resort()

    def _resizeColumns(self, index):
        """Resizes the view when items get expanded or collapsed"""
        del index   # unused argument
        # rowCount = self.model().rowCount()
        self.header().setStretchLastSection(True)

        width = max(100, self.sizeHintForColumn(0))
        self.header().resizeSection(0, width)

    def _resort(self):
        """Re-sorts the tree"""
        self.model().sort(self.header().sortIndicatorSection(),
                          self.header().sortIndicatorOrder())

    def mouseDoubleClickEvent(self, mouseEvent):
        """Reimplemented to disable expanding/collapsing of items on dbl click.

        Instead the double-clicked entry is opened.
        """
        index = self.indexAt(mouseEvent.pos())
        if not index.isValid():
            return

        item = self.model().item(index)
        if item.itemType in [GlobalsItemType,
                             ImportsItemType, FunctionsItemType,
                             ClassesItemType, StaticAttributesItemType,
                             InstanceAttributesItemType,
                             DirectoryItemType, SysPathItemType]:
            QTreeView.mouseDoubleClickEvent(self, mouseEvent)
        else:
            self.openItem(item)

    def openSelectedItem(self):
        """Triggers when an item is clicked or double clicked"""
        item = self.model().item(self.currentIndex())
        self.openItem(item)

    def highlightItem(self, path):
        """Highlights an item which matches the given path"""
        # Find the top level item first
        startItem = None
        srcModel = self.model().sourceModel()
        for treeItem in srcModel.rootItem.childItems:
            itemPath = treeItem.getPath()
            if itemPath != "" and path.startswith(itemPath):
                startItem = treeItem
                break
        if startItem is None:
            return False

        if not os.path.exists(path):
            return False

        # Here: the start item has been found and the file exists for sure
        index = srcModel.buildIndex(startItem.getRowPath())
        self.setExpanded(self.model().mapFromSource(index), True)

        parts = path[len(itemPath):].split(os.path.sep)
        dirs = parts[:-1]
        fName = parts[-1]

        for dirName in dirs:
            # find the dirName in the item and make it the current item
            found = False
            for treeItem in startItem.childItems:
                if str(treeItem.data(0)) == dirName:
                    startItem = treeItem
                    found = True
                    # Need to expand regardless it is populated or not because
                    # a dir could be populated but not expanded.
                    index = srcModel.buildIndex(startItem.getRowPath())
                    self.setExpanded(self.model().mapFromSource(index), True)
                    break
            if found:
                continue
            return False

        # Here: all the dirs have been found and they are expanded
        if fName == '':
            # It was a directory item, so there is no need to highlight,
            # it was just expanding dirs request.
            return False

        for treeItem in startItem.childItems:
            if str(treeItem.data(0)) == fName:
                # Found the item to highlight
                index = srcModel.buildIndex(treeItem.getRowPath())
                self.setCurrentIndex(self.model().mapFromSource(index))
                self.setFocus()
                return True

        return False

    def getExpanded(self):
        """Provides a list of paths which are currently expanded"""
        expandedPaths = []
        srcModel = self.model().sourceModel()
        self.__getExpanded(srcModel, srcModel.rootItem, expandedPaths)
        expandedPaths.sort()
        return expandedPaths

    def __getExpanded(self, srcModel, item, paths):
        """Provides a list of the expanded dirs in the files tree"""
        for treeItem in item.childItems:
            if treeItem.itemType not in [DirectoryItemType, SysPathItemType]:
                continue
            index = srcModel.buildIndex(treeItem.getRowPath())
            if self.isExpanded(self.model().mapFromSource(index)):
                paths.append(treeItem.getPath())
                self.__getExpanded(srcModel, treeItem, paths)

    @staticmethod
    def openItem(item):
        """Handles the case when an item is activated"""
        if item.itemType in [GlobalsItemType,
                             ImportsItemType, FunctionsItemType,
                             ClassesItemType, StaticAttributesItemType,
                             InstanceAttributesItemType,
                             DirectoryItemType, SysPathItemType]:
            return

        if item.itemType == FileItemType:
            if item.fileType is None:
                return
            if 'broken-symlink' in item.fileType:
                return

            itemPath = item.getPath()
            if not os.path.exists(itemPath):
                logging.error("Cannot open " + itemPath)
                return

            if os.path.islink(itemPath):
                # Convert it to the real path and the decide what to do
                itemPath = os.path.realpath(itemPath)
                # The type may differ...
                itemMime, _, _ = getFileProperties(itemPath)
            else:
                # The intermediate directory could be a link, so use the real
                # path
                itemPath = os.path.realpath(itemPath)
                itemMime = item.fileType

            GlobalData().mainWindow.openFileByType(itemMime, itemPath, -1)
            return

        if item.itemType in [CodingItemType, ImportItemType, FunctionItemType,
                             ClassItemType, DecoratorItemType,
                             AttributeItemType, GlobalItemType,
                             ImportWhatItemType]:
            GlobalData().mainWindow.openFile(os.path.realpath(item.getPath()),
                                             item.sourceObj.line)

    def copyToClipboard(self):
        """Copies the path to the file where the element is to the clipboard"""
        item = self.model().item(self.currentIndex())
        path = item.getPath()
        QApplication.clipboard().setText(path)

    @staticmethod
    def showParsingErrors(path):
        """Fires the show errors dialog window"""
        try:
            dialog = ParserErrorsDialog(path)
            dialog.exec_()
        except Exception as ex:
            logging.error(str(ex))

    def findInDirectory(self):
        """Find in directory popup menu handler"""
        index = self.currentIndex()
        searchDir = self.model().item(index).getPath()

        dlg = FindInFilesDialog(FindInFilesDialog.IN_DIRECTORY, "", searchDir)
        dlg.exec_()
        if dlg.searchResults:
            GlobalData().mainWindow.displayFindInFiles(dlg.searchRegexp,
                                                       dlg.searchResults)

    def selectionChanged(self, selected, deselected):
        """Triggered when the selection changed"""
        QTreeView.selectionChanged(self, selected, deselected)
        indexesList = selected.indexes()
        if indexesList:
            self.sigFirstSelectedItem.emit(indexesList[0])
        else:
            self.sigFirstSelectedItem.emit(QModelIndex())

    def _onFSChanged(self, items):
        """Triggered when the project files set has been changed"""
        itemsToDel = []
        itemsToAdd = []
        for item in items:
            item = str(item)
            if item.startswith('-'):
                itemsToDel.append(item[1:])
            else:
                itemsToAdd.append(item[1:])
        itemsToDel.sort()
        itemsToAdd.sort()

        # It is important that items are deleted first and then new are added!
        for item in itemsToDel:
            dirname, basename = self._splitPath(item)

            # For all root items
            for treeItem in self.model().sourceModel().rootItem.childItems:
                self._delFromTree(treeItem, dirname, basename)

        for item in itemsToAdd:
            dirname, basename = self._splitPath(item)

            # For all root items
            for treeItem in self.model().sourceModel().rootItem.childItems:
                self.__addToTree(treeItem, item, dirname, basename)

        self.layoutDisplay()

    def __addToTree(self, treeItem, item, dirname, basename):
        """Recursive function which adds an item to the displayed tree"""
        # treeItem is always of the directory type
        if not treeItem.populated:
            return

        srcModel = self.model().sourceModel()
        treePath = treeItem.getPath()
        if treePath != "":
            # Guard for the sys.path item
            treePath = os.path.realpath(treePath) + os.path.sep
        if treePath == dirname:
            # Need to add an item but only if there is no this item already!
            foundInChildren = False
            for i in treeItem.childItems:
                if basename == i.data(0):
                    foundInChildren = True
                    break

            if not foundInChildren:
                if item.endswith(os.path.sep):
                    newItem = TreeViewDirectoryItem(
                        treeItem, treeItem.getPath() + basename, False)
                else:
                    newItem = TreeViewFileItem(treeItem,
                                               treeItem.getPath() + basename)
                parentIndex = srcModel.buildIndex(treeItem.getRowPath())
                srcModel.addItem(newItem, parentIndex)

        for i in treeItem.childItems:
            if i.itemType == DirectoryItemType:
                self.__addToTree(i, item, dirname, basename)
            elif i.isLink:
                # Check if it was a broken link to the newly appeared item
                if os.path.realpath(i.getPath()) == dirname + basename:
                    # Update the link status
                    i.updateLinkStatus(i.getPath())
                    self._signalItemUpdated(i)

    def _delFromTree(self, treeItem, dirname, basename):
        """Recursive function which deletes an item from the displayed tree"""
        # treeItem is always of the directory type
        srcModel = self.model().sourceModel()

        d_dirname, d_basename = self._splitPath(treeItem.getPath())
        if d_dirname == dirname and d_basename == basename:
            index = srcModel.buildIndex(treeItem.getRowPath())
            srcModel.beginRemoveRows(index.parent(), index.row(), index.row())
            treeItem.parentItem.removeChild(treeItem)
            srcModel.endRemoveRows()
            return

        if treeItem.isLink:
            # Link to a directory
            if os.path.realpath(treeItem.getPath()) == dirname + basename:
                # Broken link now
                treeItem.updateStatus()
                self._signalItemUpdated(treeItem)
                return

        # Walk the directory items
        for i in treeItem.childItems:
            if i.itemType == DirectoryItemType:
                # directory
                self._delFromTree(i, dirname, basename)
            else:
                # file
                if i.isLink:
                    l_dirname, l_basename = self._splitPath(i.getPath())
                    if dirname == l_dirname and basename == l_basename:
                        index = srcModel.buildIndex(i.getRowPath())
                        srcModel.beginRemoveRows(index.parent(),
                                                 index.row(), index.row())
                        i.parentItem.removeChild(i)
                        srcModel.endRemoveRows()
                    elif os.path.realpath(i.getPath()) == dirname + basename:
                        i.updateLinkStatus(i.getPath())
                        self._signalItemUpdated(i)
                else:
                    # Regular final file
                    if os.path.realpath(i.getPath()) == dirname + basename:
                        index = srcModel.buildIndex(i.getRowPath())
                        srcModel.beginRemoveRows(index.parent(),
                                                 index.row(), index.row())
                        i.parentItem.removeChild(i)
                        srcModel.endRemoveRows()

    @staticmethod
    def _splitPath(path):
        """Provides the dirname and the base name"""
        if path.endswith(os.path.sep):
            # directory
            dirname = os.path.realpath(os.path.dirname(path[:-1])) + \
                      os.path.sep
            basename = os.path.basename(path[:-1])
        else:
            dirname = os.path.realpath(os.path.dirname(path)) + os.path.sep
            basename = os.path.basename(path)
        return dirname, basename

    def onFileUpdated(self, fileName, uuid):
        """Triggered when the file is updated"""
        del uuid    # unused argument
        mime, icon, _ = getFileProperties(fileName)
        if isPythonMime(mime):
            path = os.path.realpath(fileName)
            info = GlobalData().briefModinfoCache.get(path)
            if info.isOK:
                icon = getIcon('filepython.png')
            else:
                icon = getIcon('filepythonbroken.png')

            # For all root items
            for treeItem in self.model().sourceModel().rootItem.childItems:
                self.__walkTreeAndUpdate(treeItem, path, mime, icon, info)
        elif isCDMProjectMime(mime):
            path = os.path.realpath(fileName)
            # For all root items
            for treeItem in self.model().sourceModel().rootItem.childItems:
                self.__walkTreeAndUpdate(treeItem, path, mime, None, None)
        elif fileName.endswith(".cgi"):
            path = os.path.realpath(fileName)

            # For all root items
            for treeItem in self.model().sourceModel().rootItem.childItems:
                self.__walkTreeAndUpdate(treeItem, path, mime, icon, None)

    def onFileTypeChanged(self, fileName, uuid, newMime):
        """Triggered when the file type is changed"""
        del uuid        # unused argument
        del newMime     # unused argument
        mime, icon, _ = getFileProperties(fileName)
        path = os.path.realpath(fileName)
        for treeItem in self.model().sourceModel().rootItem.childItems:
            self.__walkTreeAndUpdate(treeItem, path, mime, icon, None)

    def _signalItemUpdated(self, treeItem):
        """Emits a signal that an item is updated"""
        srcModel = self.model().sourceModel()
        index = srcModel.buildIndex(treeItem.getRowPath())
        srcModel.dataChanged.emit(index, index)

    def __removeTreeItem(self, treeItem):
        """Removes the given item"""
        srcModel = self.model().sourceModel()
        index = srcModel.buildIndex(treeItem.getRowPath())
        srcModel.beginRemoveRows(index.parent(), index.row(), index.row())
        treeItem.parentItem.removeChild(treeItem)
        srcModel.endRemoveRows()

    def __addTreeItem(self, treeItem, newItem):
        """Adds the given item"""
        srcModel = self.model().sourceModel()
        parentIndex = srcModel.buildIndex(treeItem.getRowPath())
        srcModel.addItem(newItem, parentIndex)
        self._resort()

    def __walkTreeAndUpdate(self, treeItem, path, mime, icon, info):
        """Recursively walks the tree items and updates the icon"""
        if treeItem.itemType in [DirectoryItemType, SysPathItemType]:
            for i in treeItem.childItems:
                if i.itemType in [DirectoryItemType,
                                  SysPathItemType, FileItemType]:
                    self.__walkTreeAndUpdate(i, path, mime, icon, info)

        if treeItem.itemType == FileItemType:
            if path == os.path.realpath(treeItem.getPath()):
                if isPythonMime(mime) and info:
                    # Update icon
                    treeItem.setIcon(icon)
                    if info.docstring is None:
                        treeItem.toolTip = ""
                    else:
                        treeItem.toolTip = info.docstring.text
                    treeItem.parsingErrors = not info.isOK

                    self._signalItemUpdated(treeItem)

                    # Update content if populated
                    self.updateFileItem(treeItem, info)
                elif isCDMProjectMime(mime):
                    # Tooltip update only
                    treeItem.toolTip = getProjectFileTooltip(path)
                    self._signalItemUpdated(treeItem)
                elif path.endswith(".cgi"):
                    # It can only happened when python CGI is not a python any
                    # more. So display it as a general file.
                    # The case when a cgi became a python file is covered in
                    # the first branch of this if statement.
                    treeItem.setIcon(icon)
                    treeItem.toolTip = ""
                    self._signalItemUpdated(treeItem)

                    # Remove child items if so
                    while treeItem.childItems:
                        self.__removeTreeItem(treeItem.childItems[0])
                else:
                    treeItem.setIcon(icon)
                    treeItem.toolTip = ""
                    self._signalItemUpdated(treeItem)

    def updateFileItem(self, treeItem, info):
        """Updates the file item recursively"""
        hadCoding = False
        hadGlobals = False
        hadImports = False
        hadFunctions = False
        hadClasses = False
        itemsToRemove = []
        for fileChildItem in treeItem.childItems:
            if fileChildItem.itemType == CodingItemType:
                hadCoding = True
                if info.encoding is None:
                    itemsToRemove.append(fileChildItem)
                else:
                    fileChildItem.updateData(info.encoding)
                    self._signalItemUpdated(fileChildItem)
                continue
            elif fileChildItem.itemType == GlobalsItemType:
                hadGlobals = True
                if info.globals:
                    fileChildItem.updateData(info)
                    self.__updateGlobalsItem(fileChildItem, info.globals)
                else:
                    itemsToRemove.append(fileChildItem)
                continue
            elif fileChildItem.itemType == ImportsItemType:
                hadImports = True
                if info.imports:
                    fileChildItem.updateData(info)
                    self.__updateImportsItem(fileChildItem, info.imports)
                else:
                    itemsToRemove.append(fileChildItem)
                continue
            elif fileChildItem.itemType == FunctionsItemType:
                hadFunctions = True
                if info.functions:
                    fileChildItem.updateData(info)
                    self.model().sourceModel().updateFunctionsItem(
                        fileChildItem, info.functions)
                    self._resort()
                else:
                    itemsToRemove.append(fileChildItem)
                continue
            elif fileChildItem.itemType == ClassesItemType:
                hadClasses = True
                if info.classes:
                    fileChildItem.updateData(info)
                    self.model().sourceModel().updateClassesItem(fileChildItem,
                                                                 info.classes)
                    self._resort()
                else:
                    itemsToRemove.append(fileChildItem)
                continue

        for item in itemsToRemove:
            self.__removeTreeItem(item)

        if not hadCoding and treeItem.populated and \
           info.encoding is not None:
            # Coding item appeared, so we need to add it
            newItem = TreeViewCodingItem(treeItem, info.encoding)
            self.__addTreeItem(treeItem, newItem)

        if not hadGlobals and treeItem.populated and info.globals:
            # Globals item appeared, so we need to add it
            newItem = TreeViewGlobalsItem(treeItem, info)
            self.__addTreeItem(treeItem, newItem)

        if not hadImports and treeItem.populated and info.imports:
            # Imports item appeared, so we need to add it
            newItem = TreeViewImportsItem(treeItem, info)
            self.__addTreeItem(treeItem, newItem)

        if not hadFunctions and treeItem.populated and info.functions:
            # Functions item appeared, so we need to add it
            newItem = TreeViewFunctionsItem(treeItem, info)
            self.__addTreeItem(treeItem, newItem)

        if not hadClasses and treeItem.populated and info.classes:
            # Classes item appeared, so we need to add it
            newItem = TreeViewClassesItem(treeItem, info)
            self.__addTreeItem(treeItem, newItem)

    def __updateGlobalsItem(self, treeItem, globalsObj):
        """Updates globals item"""
        if not treeItem.populated:
            return

        existingGlobals = []
        itemsToRemove = []
        for globalItem in treeItem.childItems:
            name = globalItem.data(0)
            found = False
            for glob in globalsObj:
                if glob.name == name:
                    found = True
                    existingGlobals.append(name)
                    globalItem.updateData(glob)
                    # No need to send the update signal because the name is
                    # still the same
                    break
            if not found:
                # Disappeared item
                itemsToRemove.append(globalItem)

        for item in itemsToRemove:
            self.__removeTreeItem(item)

        # Add those which have been introduced
        for glob in globalsObj:
            if glob.name not in existingGlobals:
                newItem = TreeViewGlobalItem(treeItem, glob)
                self.__addTreeItem(treeItem, newItem)

    def __updateImportsItem(self, treeItem, importsObj):
        """Updates imports item"""
        if not treeItem.populated:
            return

        # Need to update item by item. There could be 2 import items with
        # the same name, so this stuff of a copied list.
        importsCopy = list(importsObj)
        itemsToRemove = []
        for importItem in treeItem.childItems:
            name = importItem.data(0)
            found = False
            for index in range(len(importsCopy)):
                if importsCopy[index].getDisplayName() == name:
                    found = True
                    importItem.updateData(importsCopy[index])
                    # No need to send the update signal because the name is
                    # still the same, but need to update the importwhat items
                    # if so
                    self.__updateSingleImportItem(importItem,
                                                  importsCopy[index])
                    del importsCopy[index]
                    break
            if not found:
                # Disappeared item
                itemsToRemove.append(importItem)

        for item in itemsToRemove:
            self.__removeTreeItem(item)

        # Add those which have been introduced
        for item in importsCopy:
            newItem = TreeViewImportItem(treeItem, item)
            self.__addTreeItem(treeItem, newItem)

    def __updateSingleImportItem(self, treeItem, importObject):
        """Updates single import item, i.e. importWhat"""
        if not treeItem.populated:
            return
        importWhatCopy = list(importObject.what)
        itemsToRemove = []
        for importWhatItem in treeItem.childItems:
            name = importWhatItem.data(0)
            found = False
            for index in range(len(importWhatCopy)):
                if importWhatCopy[index].getDisplayName() == name:
                    found = True
                    importWhatItem.updateData(importWhatCopy[index])
                    # No need to send the update signal because the name is
                    # still the same
                    del importWhatCopy[index]
                    break
            if not found:
                # Disappeared item
                itemsToRemove.append(importWhatItem)

        for item in itemsToRemove:
            self.__removeTreeItem(item)

        # Add those which have been introduced
        for item in importWhatCopy:
            newItem = TreeViewWhatItem(treeItem, item)
            self.__addTreeItem(treeItem, newItem)

    def onDebugMode(self, newState):
        """Memorizes the current debug state"""
        self.__debugMode = newState



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""File system browser model"""

from utils.globals import GlobalData
from utils.project import CodimensionProject
from utils.settings import Settings
from .qt import QModelIndex, QDir
from .viewitems import TreeViewDirectoryItem, TreeViewSysPathItem
from .browsermodelbase import BrowserModelBase


class FileSystemBrowserModel(BrowserModelBase):

    """Class implementing the file system browser model"""

    def __init__(self, parent=None):
        BrowserModelBase.__init__(self, "Name", parent)
        self.setTooltips(Settings()['projectTooltips'])

        self.projectTopLevelDirs = []
        self.populateModel()

        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)

    def populateModel(self):
        """Populates the browser model"""
        self.clear()
        self.addItem(TreeViewSysPathItem(self.rootItem))
        self.addItem(TreeViewDirectoryItem(self.rootItem, QDir.homePath()))
        for dname in QDir.drives():
            self.addItem(TreeViewDirectoryItem(self.rootItem,
                                               dname.absoluteFilePath()))
        self.__populateProjectTopLevelDirs()

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what != CodimensionProject.CompleteProject:
            return

        self.__removeProjectTopLevelDirs()      # Remove the previous set
        self.__populateProjectTopLevelDirs()    # Populate the new one

    def __populateProjectTopLevelDirs(self):
        """Populates the project specific top level dirs"""
        for dname in GlobalData().project.topLevelDirs:
            cnt = self.rootItem.childCount()
            self.beginInsertRows(QModelIndex(), cnt, cnt)
            self._addItem(TreeViewDirectoryItem(self.rootItem, dname),
                          self.rootItem)
            self.endInsertRows()
            self.projectTopLevelDirs.append(dname)

    def __removeProjectTopLevelDirs(self):
        """Removes all the project top level dirs"""
        while len(self.projectTopLevelDirs) > 0:
            # Find the row index in the children list of the root item
            idx = 0
            for item in self.rootItem.childItems:
                if item.getPath() == self.projectTopLevelDirs[0]:
                    index = self.index(idx, 0)
                    self.removeTopLevelDir(index)
                    break
                idx += 1

    def addTopLevelDir(self, dirname):
        """Adds a new toplevel directory"""
        itm = TreeViewDirectoryItem(self.rootItem, dirname)
        self.addItem(itm)
        self.projectTopLevelDirs.append(dirname)

    def removeTopLevelDir(self, index):
        """Removes a toplevel directory"""
        if index.isValid():
            item = index.internalPointer()
            self.beginRemoveRows(index.parent(), index.row(), index.row())
            self.rootItem.removeChild(item)
            self.endRemoveRows()

            self.projectTopLevelDirs.remove(item.getPath())



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""ItemDelegate which helps changing the standard row height"""


from .qt import (QItemDelegate, QStyledItemDelegate, QStyleOptionViewItem,
                 QStyle)


class ChangingHeightItemDelegate(QItemDelegate):

    """Helper class to set rows height"""

    def __init__(self, delta):
        QItemDelegate.__init__(self)
        self.delta = delta
        self.lastHeight = 0

    def sizeHint(self, option, index):
        """Returns the size hint in which the only height matters"""
        origSize = QItemDelegate.sizeHint(self, option, index)
        self.lastHeight = origSize.height() + self.delta
        origSize.setHeight(self.lastHeight)
        return origSize


class NoOutlineDelegate(QStyledItemDelegate):

    """Hides the dotted line outline around tree view cells"""

    def __init__(self):
        QStyledItemDelegate.__init__(self)

    def paint(self, painter, option, index):
        """Hides the dotted outline"""
        itemOption = QStyleOptionViewItem(option)
        if itemOption.state & QStyle.State_HasFocus != 0:
            itemOption.state = itemOption.state & ~QStyle.State_HasFocus
        QStyledItemDelegate.paint(self, painter, itemOption, index)


class NoOutlineHeightDelegate(QStyledItemDelegate):

    """Changes the raw height and removes the dotted cells outline"""

    def __init__(self, delta):
        QStyledItemDelegate.__init__(self)
        self.delta = delta
        self.lastHeight = 0

    def paint(self, painter, option, index):
        """Hides the dotted outline"""
        itemOption = QStyleOptionViewItem(option)
        if itemOption.state & QStyle.State_HasFocus != 0:
            itemOption.state = itemOption.state & ~QStyle.State_HasFocus
        QStyledItemDelegate.paint(self, painter, itemOption, index)

    def sizeHint(self, option, index):
        """Returns the size hint in which the only height matters"""
        origSize = QStyledItemDelegate.sizeHint(self, option, index)
        self.lastHeight = origSize.height() + self.delta
        origSize.setHeight(self.lastHeight)
        return origSize



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2011-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""The debugger namespace viewer implementation"""


from ui.qt import (Qt, QRegExp, QAbstractItemView, QHeaderView, QTreeWidget)
from ui.itemdelegates import NoOutlineHeightDelegate
from utils.settings import Settings
from .variableitems import (VariableItem, SpecialVariableItem,
                            ArrayElementVariableItem,
                            SpecialArrayElementVariableItem,
                            INDICATORS)
from .varfilters import VARIABLE_FILTERS
from .viewvariable import ViewVariableDialog
from .client.protocol_cdm_dbg import VAR_TYPE_DISP_STRINGS


INDICATOR_PATTERN = "|".join([QRegExp.escape(indicator)
                              for indicator in INDICATORS])
RX_CLASS = QRegExp('<.*(instance|object) at 0x.*>')
RX_CLASS2 = QRegExp('class .*')
RX_CLASS3 = QRegExp('<class .* at 0x.*>')
DVAR_RX_CLASS1 = QRegExp(
    r'<.*(instance|object) at 0x.*>({0})'.format(INDICATOR_PATTERN))
DVAR_RX_CLASS2 = QRegExp(
    r'<class .* at 0x.*>({0})'.format(INDICATOR_PATTERN))
DVAR_RX_ARRAY_ELEMENT = QRegExp(r'^\d+$')
DVAR_RX_SPECIAL_ARRAY_ELEMENT = QRegExp(
    r'^\d+({0})$'.format(INDICATOR_PATTERN))
RX_NONPRINTABLE = QRegExp(r"""(\\x\d\d)+""")

TYPE_TO_INDICATORS = {
    # Python types
    'list': '[]',
    'tuple': '()',
    'dict': '{:}',
    'set': '{}',
    'frozenset': '{}'}


class VariablesBrowser(QTreeWidget):

    """Variables browser implementation"""

    def __init__(self, debugger, parent=None):
        QTreeWidget.__init__(self, parent)

        self.setRootIsDecorated(True)
        self.setAlternatingRowColors(True)
        self.setUniformRowHeights(True)
        self.setItemDelegate(NoOutlineHeightDelegate(4))

        self.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.setHeaderLabels(["Name", "Value", "Type"])
        header = self.header()
        header.setSortIndicator(0, Qt.AscendingOrder)
        header.setSortIndicatorShown(True)
        header.setSectionsClickable(True)
        header.setStretchLastSection(True)

        self.itemExpanded.connect(self.__expandItemSignal)
        self.itemCollapsed.connect(self.collapseItem)

        self.resortEnabled = True
        self.openItems = []
        self.framenr = 0
        self.__debugger = debugger

        # Ugly filtering support
        self.__filterIsSet = False

        self.setSortingEnabled(True)

    def scrollTo(self, index, hint=QAbstractItemView.EnsureVisible):
        """ Disables horizontal scrolling when a row is clicked.

        Found here: http://qt-project.org/faq/answer/
                           how_can_i_disable_autoscroll_when
                           _selecting_a_partially_displayed_column_in
        """
        oldValue = self.horizontalScrollBar().value()
        QTreeWidget.scrollTo(self, index, hint)
        self.horizontalScrollBar().setValue(oldValue)

    def __findItem(self, slist, column, node=None):
        """Searches for an item.

        It is used to find a specific item in column,
        that is a child of node. If node is None, a child of the
        QTreeWidget is searched.
        """
        if node is None:
            count = self.topLevelItemCount()
        else:
            count = node.childCount()

        if column == 0:
            searchStr = VariableItem.extractId(slist[0])[0]
        else:
            searchStr = slist[0]

        for index in range(count):
            if node is None:
                item = self.topLevelItem(index)
            else:
                item = node.child(index)

            if item.text(column) == searchStr:
                if len(slist) > 1:
                    item = self.__findItem(slist[1:], column, item)
                return item
        return None

    def __clearScopeVariables(self, areGlobals):
        """Removes those variables which belong to the specified frame"""
        count = self.topLevelItemCount()
        for index in range(count - 1, -1, -1):
            item = self.topLevelItem(index)
            if item.isGlobal() == areGlobals:
                self.takeTopLevelItem(index)

    def showVariables(self, areGlobals, vlist, frmnr):
        """Shows variables list"""
        self.current = self.currentItem()
        if self.current:
            self.curpathlist = self.__buildTreePath(self.current)
        self.__clearScopeVariables(areGlobals)
        self.__scrollToItem = None
        self.framenr = frmnr

        if len(vlist):
            self.resortEnabled = False
            self.setSortingEnabled(False)
            for (var, vtype, value) in vlist:
                item = self.__addItem(None, areGlobals, vtype, var, value)
                item.setHidden(not self.__variableToShow(item.getName(),
                                                         item.isGlobal(),
                                                         item.getType()))

            # reexpand tree
            openItems = sorted(self.openItems[:])
            self.openItems = []
            for itemPath in openItems:
                itm = self.__findItem(itemPath, 0)
                if itm is not None:
                    self.expandItem(itm)
                else:
                    self.openItems.append(itemPath)

            if self.current:
                citm = self.__findItem(self.curpathlist, 0)
                if citm:
                    self.setCurrentItem(citm)
                    citm.setSelected(True)
                    self.scrollToItem(citm, QAbstractItemView.PositionAtTop)
                    self.current = None

            self.__resizeSections()

            self.resortEnabled = True
            self.setSortingEnabled(True)
            self.__resort()

    def __resizeSections(self):
        """Resizes the variable sections"""
        if self.topLevelItemCount() == 0:
            return

        header = self.header()
        nameSectionSize = header.sectionSize(0)
        header.resizeSections(QHeaderView.ResizeToContents)
        if header.sectionSize(0) < nameSectionSize:
            header.resizeSection(0, nameSectionSize)

    def showVariable(self, isGlobal, vlist):
        """Shows variables in a list"""
        # vlist the list of subitems to be displayed. The first element gives
        # the path of the parent variable. Each other listentry is a tuple of
        # three values:
        #   the variable name (string)
        #   the variables type (string)
        #   the variables value (string)

        resortEnabled = self.resortEnabled
        self.resortEnabled = False
        self.setSortingEnabled(False)

        if self.current is None:
            self.current = self.currentItem()
            if self.current:
                self.curpathlist = self.__buildTreePath(self.current)

        if vlist:
            item = self.__findItem(vlist[0], 0)
            for var, vtype, value in vlist[1:]:
                newItem = self.__addItem(item, isGlobal, vtype, var, value)
                newItem.setHidden(not self.__variableToShow(newItem.getName(),
                                                            newItem.isGlobal(),
                                                            newItem.getType()))

        # reexpand tree
        openItems = sorted(self.openItems[:])
        self.openItems = []
        for itemPath in openItems:
            item = self.__findItem(itemPath, 0)
            if item is not None and not item.isExpanded():
                if item.populated:
                    self.blockSignals(True)
                    item.setExpanded(True)
                    self.blockSignals(False)
                else:
                    self.expandItem(item)
        self.openItems = openItems[:]

        if self.current:
            citm = self.__findItem(self.curpathlist, 0)
            if citm:
                self.setCurrentItem(citm)
                citm.setSelected(True)
                if self.__scrollToItem:
                    self.scrollToItem(self.__scrollToItem,
                                      QAbstractItemView.PositionAtTop)
                else:
                    self.scrollToItem(citm, QAbstractItemView.PositionAtTop)
                self.current = None
        elif self.__scrollToItem:
            self.scrollToItem(self.__scrollToItem,
                              QAbstractItemView.PositionAtTop)

        self.__resizeSections()

        self.resortEnabled = resortEnabled
        if self.resortEnabled:
            self.setSortingEnabled(True)
        self.__resort()

    def __generateItem(self, parentItem, isGlobal,
                       varName, varValue, varType, isSpecial=False):
        """Generates an appropriate variable item"""
        if isSpecial:
            if DVAR_RX_CLASS1.exactMatch(varName) or \
               DVAR_RX_CLASS2.exactMatch(varName):
                isSpecial = False

        if RX_CLASS2.exactMatch(varType):
            return SpecialVariableItem(parentItem, self.__debugger, isGlobal,
                                       varName, varValue, varType[7:-1],
                                       self.framenr)

        elif varType != "void *" and \
             (RX_CLASS.exactMatch(varValue) or
              RX_CLASS3.exactMatch(varValue) or
              isSpecial):
            if DVAR_RX_SPECIAL_ARRAY_ELEMENT.exactMatch(varName):
                return SpecialArrayElementVariableItem(parentItem,
                                                       self.__debugger,
                                                       isGlobal,
                                                       varName, varValue,
                                                       varType,
                                                       self.framenr)
            return SpecialVariableItem(parentItem, self.__debugger, isGlobal,
                                       varName, varValue, varType,
                                       self.framenr)
        elif varType in ["numpy.ndarray", "django.MultiValueDict",
                         "array.array"]:
            return SpecialVariableItem(parentItem, self.__debugger, isGlobal,
                                       varName, "{0} items".format(varValue),
                                       varType, self.framenr)
        else:
            if DVAR_RX_ARRAY_ELEMENT.exactMatch(varName):
                return ArrayElementVariableItem(parentItem, isGlobal,
                                                varName, varValue, varType)
            return VariableItem(parentItem, isGlobal,
                                varName, varValue, varType)

        print("WARNING: Reached the end without forming a variable!")

    def __getDisplayType(self, varType):
        """Provides a variable type for display purpose"""
        try:
            dvtype = VAR_TYPE_DISP_STRINGS[varType]
        except KeyError:
            if varType == 'classobj':
                dvtype = VAR_TYPE_DISP_STRINGS['instance']
            else:
                dvtype = varType
        return dvtype

    def __addItem(self, parentItem, isGlobal, varType, varName, varValue):
        """Adds a new item to the children of the parentItem"""
        if parentItem is None:
            parentItem = self

        try:
            varName = '{0}{1}'.format(varName, TYPE_TO_INDICATORS[varType])
        except KeyError:
            pass

        displayType = self.__getDisplayType(varType)
        if varType in ['list', 'tuple', 'dict', 'set', 'frozenset']:
            return self.__generateItem(parentItem, isGlobal,
                                       varName, str(varValue) + " item(s)",
                                       displayType,
                                       True)
        if varType in ['unicode', 'str']:
            if RX_NONPRINTABLE.indexIn(varValue) != -1:
                stringValue = varValue
            else:
                try:
                    stringValue = eval(varValue)
                    displayType += " (chars: " + str(len(stringValue)) + ")"
                except:
                    stringValue = varValue
            return self.__generateItem(parentItem, isGlobal,
                                       varName, stringValue,
                                       displayType)

        return self.__generateItem(parentItem, isGlobal,
                                   varName, varValue, displayType)

    def mouseDoubleClickEvent(self, mouseEvent):
        """Handles the mouse double click event"""
        item = self.itemAt(mouseEvent.pos())
        if item is None:
            return

        childCount = item.childCount()
        if childCount == 0:
            # Show the dialog
            dlg = ViewVariableDialog(self.__getQualifiedName(item),
                                     item.getType(), item.getValue(),
                                     item.isGlobal())
            dlg.exec_()
        else:
            QTreeWidget.mouseDoubleClickEvent(self, mouseEvent)

    def __getQualifiedName(self, item):
        """Provides a fully qualified name"""
        name = item.getName()
        if name[-2:] in ['[]', '{}', '()']:
            name = name[:-2]

        par = item.parent()
        nlist = [name]
        # build up the fully qualified name
        while par is not None:
            pname = par.getName()
            if pname[-2:] in ['[]', '{}', '()']:
                if nlist[0].endswith("."):
                    nlist[0] = '[%s].' % nlist[0][:-1]
                else:
                    nlist[0] = '[%s]' % nlist[0]
                nlist.insert(0, pname[:-2])
            else:
                nlist.insert(0, '%s.' % pname)
            par = par.parent()

        name = ''.join(nlist)
        return name

    def __buildTreePath(self, item):
        """Builds up a path from the top to the given item"""
        name = item.text(0)
        pathList = [name]

        parent = item.parent()
        # build up a path from the top to the item
        while parent is not None:
            parentVariableName = parent.text(0)
            pathList.insert(0, parentVariableName)
            parent = parent.parent()

        return pathList[:]

    def __expandItemSignal(self, parentItem):
        """Handles the expanded signal"""
        self.expandItem(parentItem)
        self.__scrollToItem = parentItem

    def expandItem(self, parentItem):
        """Handles the expanded signal"""
        pathList = self.__buildTreePath(parentItem)
        self.openItems.append(pathList)
        if parentItem.populated:
            return

        try:
            parentItem.expand()
            self.__resort()
        except AttributeError:
            QTreeWidget.expandItem(self, parentItem)

    def collapseItem(self, parentItem):
        """Handles the collapsed signal"""
        pathList = self.__buildTreePath(parentItem)
        self.openItems.remove(pathList)

        try:
            parentItem.collapse()
        except AttributeError:
            QTreeWidget.collapseItem(self, parentItem)

    def __resort(self):
        """Resorts the tree"""
        if self.resortEnabled:
            self.sortItems(self.sortColumn(),
                           self.header().sortIndicatorOrder())

    def getShownAndTotalCounts(self):
        """Provides the total number of variables and currently shown"""
        total = self.topLevelItemCount()
        shownCount = 0
        for index in range(total):
            if not self.topLevelItem(index).isHidden():
                shownCount += 1
        return shownCount, total

    def clear(self):
        """Resets everything"""
        self.resortEnabled = True
        self.openItems = []
        self.framenr = 0

        QTreeWidget.clear(self)

    def filterChanged(self):
        """Sets the new filter"""
        self.__filterIsSet = False
        for _, settingName, _ in VARIABLE_FILTERS:
            if not Settings()[settingName]:
                self.__filterIsSet = True
                break
        self.__applyFilters()

    def __applyFilters(self):
        """Re-applies the filters to the list"""
        resortEnabled = self.resortEnabled
        self.resortEnabled = False
        self.setSortingEnabled(False)

        for index in range(self.topLevelItemCount()):
            item = self.topLevelItem(index)
            toShow = self.__variableToShow(item.getName(),
                                           item.isGlobal(),
                                           item.getType())
            item.setHidden(not toShow)
            if toShow:
                self.__applyFiltersRecursively(item)

        self.__resizeSections()
        self.resortEnabled = resortEnabled
        if self.resortEnabled:
            self.setSortingEnabled(True)
        self.__resort()

    def __applyFiltersRecursively(self, item):
        """Applies the filter recursively to all the children of the item"""
        for index in range(item.childCount()):
            childItem = item.child(index)
            if not hasattr(childItem, "getName"):
                continue
            toShow = self.__variableToShow(childItem.getName(),
                                           childItem.isGlobal(),
                                           childItem.getType())
            childItem.setHidden(not toShow)
            if toShow:
                self.__applyFiltersRecursively(childItem)

    def __variableToShow(self, varName, isGlobal, varType):
        """Checks if the given item should be shown"""
        if not self.__filterIsSet:
            return True

        # Strip indicators from the name
        varName = str(varName)
        if varName.endswith("]") or \
           varName.endswith("}") or \
           varName.endswith(")"):
            varName = varName[:-2]   # Strip display purpose decor if so

        settings = Settings()
        for _, settingName, matchFunction in VARIABLE_FILTERS:
            if settings[settingName] is False:
                if matchFunction(isGlobal, varName, varType):
                    return False
        return True



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Text editor tab widget"""


import os.path
import logging
from ui.qt import (Qt, QFileInfo, QSize, pyqtSignal, QToolBar, QHBoxLayout,
                   QWidget, QAction, QMenu, QToolButton, QDialog,
                   QVBoxLayout, QSplitter)
from ui.mainwindowtabwidgetbase import MainWindowTabWidgetBase
from ui.importlist import ImportListWidget
from ui.outsidechanges import OutsideChangeWidget
from ui.spacers import ToolBarExpandingSpacer, ToolBarVSpacer
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.settings import Settings
from utils.misc import extendInstance
from utils.fileutils import isPythonMime
from utils.importutils import getImportsList, resolveImports
from diagram.importsdgm import (ImportsDiagramDialog, ImportDiagramOptions,
                                ImportsDiagramProgress)
from .flowuiwidget import FlowUIWidget
from .mdwidget import MDWidget
from .navbar import NavigationBar
from .texteditor import TextEditor


class TextEditorTabWidget(QWidget):

    """Plain text editor tab widget"""

    sigReloadRequest = pyqtSignal()
    reloadAllNonModifiedRequest = pyqtSignal()
    sigTabRunChanged = pyqtSignal(bool)

    def __init__(self, parent, debugger):
        QWidget.__init__(self, parent)

        extendInstance(self, MainWindowTabWidgetBase)
        MainWindowTabWidgetBase.__init__(self)

        self.__navigationBar = None
        self.__editor = TextEditor(self, debugger)
        self.__fileName = ""
        self.__shortName = ""

        self.__createLayout()

        self.__editor.redoAvailable.connect(self.__redoAvailable)
        self.__editor.undoAvailable.connect(self.__undoAvailable)
        self.__editor.modificationChanged.connect(self.modificationChanged)
        self.__editor.sigCFlowSyncRequested.connect(self.cflowSyncRequested)
        self.__editor.languageChanged.connect(self.__languageChanged)

        self.__diskModTime = None
        self.__diskSize = None
        self.__reloadDlgShown = False

        self.__debugMode = False

        self.__vcsStatus = None

    def onTextZoomChanged(self):
        """Triggered when a text zoom is changed"""
        self.__editor.onTextZoomChanged()

    def onFlowZoomChanged(self):
        """Triggered when a flow zoom is changed"""
        self.__flowUI.onFlowZoomChanged()

    def getNavigationBar(self):
        """Provides a reference to the navigation bar"""
        return self.__navigationBar

    def shouldAcceptFocus(self):
        """True if it can accept the focus"""
        return self.__outsideChangesBar.isHidden()

    def readFile(self, fileName):
        """Reads the text from a file"""
        self.__editor.readFile(fileName)
        self.setFileName(fileName)
        self.__editor.restoreBreakpoints()

        # Memorize the modification date
        path = os.path.realpath(fileName)
        self.__diskModTime = os.path.getmtime(path)
        self.__diskSize = os.path.getsize(path)

    def writeFile(self, fileName):
        """Writes the text to a file"""
        if self.__editor.writeFile(fileName):
            # Memorize the modification date
            path = os.path.realpath(fileName)
            self.__diskModTime = os.path.getmtime(path)
            self.__diskSize = os.path.getsize(path)
            self.setFileName(fileName)
            self.__editor.restoreBreakpoints()
            return True
        return False

    def __createLayout(self):
        """Creates the toolbar and layout"""
        # The toolbar
        self.toolbar = QToolBar(self)
        self.toolbar.setOrientation(Qt.Vertical)
        self.toolbar.setMovable(False)
        self.toolbar.setAllowedAreas(Qt.RightToolBarArea)
        self.toolbar.setIconSize(QSize(16, 16))
        self.toolbar.setFixedWidth(30)
        self.toolbar.setContentsMargins(0, 0, 0, 0)

        # Buttons
        printButton = QAction(getIcon('printer.png'), 'Print (Ctrl+P)',
                              self.toolbar)
        printButton.triggered.connect(self.__onPrint)
        printButton.setObjectName('printButton')

        printPreviewButton = QAction(getIcon('printpreview.png'),
                                     'Print preview', self.toolbar)
        printPreviewButton.triggered.connect(self.__onPrintPreview)
        printPreviewButton.setEnabled(False)
        printPreviewButton.setVisible(False)
        printPreviewButton.setObjectName('printPreviewButton')

        printSpacer = ToolBarVSpacer(self.toolbar, 8)
        printSpacer.setObjectName('printSpacer')

        # Imports diagram and its menu
        importsMenu = QMenu(self)
        self.importsDlgAct = importsMenu.addAction(
            getIcon('detailsdlg.png'), 'Fine tuned imports diagram')
        self.importsDlgAct.triggered.connect(self.onImportDgmTuned)
        self.importsDiagramButton = QToolButton(self.toolbar)
        self.importsDiagramButton.setIcon(getIcon('importsdiagram.png'))
        self.importsDiagramButton.setToolTip('Generate imports diagram')
        self.importsDiagramButton.setPopupMode(QToolButton.DelayedPopup)
        self.importsDiagramButton.setMenu(importsMenu)
        self.importsDiagramButton.setFocusPolicy(Qt.NoFocus)
        self.importsDiagramButton.clicked.connect(self.onImportDgm)
        self.importsDiagramButton.setEnabled(False)
        self.importsDiagramButton.setObjectName('importsDiagramButton')

        # Run script and its menu
        runScriptMenu = QMenu(self)
        self.runScriptDlgAct = runScriptMenu.addAction(
            getIcon('detailsdlg.png'), 'Set run/debug parameters')
        self.runScriptDlgAct.triggered.connect(self.onRunScriptDlg)
        self.runScriptButton = QToolButton(self.toolbar)
        self.runScriptButton.setIcon(getIcon('run.png'))
        self.runScriptButton.setToolTip('Run script')
        self.runScriptButton.setPopupMode(QToolButton.DelayedPopup)
        self.runScriptButton.setMenu(runScriptMenu)
        self.runScriptButton.setFocusPolicy(Qt.NoFocus)
        self.runScriptButton.clicked.connect(self.onRunScript)
        self.runScriptButton.setEnabled(False)
        self.runScriptButton.setObjectName('runScriptButton')

        # Profile script and its menu
        profileScriptMenu = QMenu(self)
        self.profileScriptDlgAct = profileScriptMenu.addAction(
            getIcon('detailsdlg.png'), 'Set profile parameters')
        self.profileScriptDlgAct.triggered.connect(self.onProfileScriptDlg)
        self.profileScriptButton = QToolButton(self.toolbar)
        self.profileScriptButton.setIcon(getIcon('profile.png'))
        self.profileScriptButton.setToolTip('Profile script')
        self.profileScriptButton.setPopupMode(QToolButton.DelayedPopup)
        self.profileScriptButton.setMenu(profileScriptMenu)
        self.profileScriptButton.setFocusPolicy(Qt.NoFocus)
        self.profileScriptButton.clicked.connect(self.onProfileScript)
        self.profileScriptButton.setEnabled(False)
        self.profileScriptButton.setObjectName('profileScriptButton')

        # Debug script and its menu
        debugScriptMenu = QMenu(self)
        self.debugScriptDlgAct = debugScriptMenu.addAction(
            getIcon('detailsdlg.png'), 'Set run/debug parameters')
        self.debugScriptDlgAct.triggered.connect(self.onDebugScriptDlg)
        self.debugScriptButton = QToolButton(self.toolbar)
        self.debugScriptButton.setIcon(getIcon('debugger.png'))
        self.debugScriptButton.setToolTip('Debug script')
        self.debugScriptButton.setPopupMode(QToolButton.DelayedPopup)
        self.debugScriptButton.setMenu(debugScriptMenu)
        self.debugScriptButton.setFocusPolicy(Qt.NoFocus)
        self.debugScriptButton.clicked.connect(self.onDebugScript)
        self.debugScriptButton.setEnabled(False)
        self.debugScriptButton.setObjectName('debugScriptButton')

        # Disassembling
        disasmScriptMenu = QMenu(self)
        disasmScriptMenu.addAction(
            getIcon(''), 'Disassembly (no optimization)',
            self.__editor._onDisasm0)
        disasmScriptMenu.addAction(
            getIcon(''), 'Disassembly (optimization level 1)',
            self.__editor._onDisasm1)
        disasmScriptMenu.addAction(
            getIcon(''), 'Disassembly (optimization level 2)',
            self.__editor._onDisasm2)
        self.disasmScriptButton = QToolButton(self.toolbar)
        self.disasmScriptButton.setIcon(getIcon('disassembly.png'))
        self.disasmScriptButton.setToolTip('Disassembly script')
        self.disasmScriptButton.setPopupMode(QToolButton.DelayedPopup)
        self.disasmScriptButton.setMenu(disasmScriptMenu)
        self.disasmScriptButton.setFocusPolicy(Qt.NoFocus)
        self.disasmScriptButton.clicked.connect(self.__editor._onDisasm0)
        self.disasmScriptButton.setEnabled(False)
        self.disasmScriptButton.setObjectName('disasmScriptButton')

        # Dead code
        self.deadCodeScriptButton = QAction(getIcon('deadcode.png'),
                                            'Find dead code', self.toolbar)
        self.deadCodeScriptButton.triggered.connect(self.__onDeadCode)
        self.deadCodeScriptButton.setEnabled(False)
        self.deadCodeScriptButton.setObjectName('deadCodeScriptButton')

        undoSpacer = ToolBarVSpacer(self.toolbar, 8)
        undoSpacer.setObjectName('undoSpacer')

        self.__undoButton = QAction(getIcon('undo.png'), 'Undo (Ctrl+Z)',
                                    self.toolbar)
        self.__undoButton.setShortcut('Ctrl+Z')
        self.__undoButton.triggered.connect(self.__editor.onUndo)
        self.__undoButton.setEnabled(False)
        self.__undoButton.setObjectName('undoButton')

        self.__redoButton = QAction(getIcon('redo.png'), 'Redo (Ctrl+Y)',
                                    self.toolbar)
        self.__redoButton.setShortcut('Ctrl+Y')
        self.__redoButton.triggered.connect(self.__editor.onRedo)
        self.__redoButton.setEnabled(False)
        self.__redoButton.setObjectName('redoButton')

        spacer = ToolBarExpandingSpacer(self.toolbar)
        spacer.setObjectName('spacer')

        self.removeTrailingSpacesButton = QAction(
            getIcon('trailingws.png'), 'Remove trailing spaces', self.toolbar)
        self.removeTrailingSpacesButton.triggered.connect(
            self.onRemoveTrailingWS)
        self.removeTrailingSpacesButton.setObjectName(
            'removeTrailingSpacesButton')
        self.expandTabsButton = QAction(
            getIcon('expandtabs.png'), 'Expand tabs (4 spaces)', self.toolbar)
        self.expandTabsButton.triggered.connect(self.onExpandTabs)
        self.expandTabsButton.setObjectName('expandTabsButton')

        # Add items to the toolbar
        self.toolbar.addAction(printPreviewButton)
        self.toolbar.addAction(printButton)
        self.toolbar.addWidget(printSpacer)
        self.toolbar.addWidget(self.importsDiagramButton)
        self.toolbar.addWidget(self.runScriptButton)
        self.toolbar.addWidget(self.profileScriptButton)
        self.toolbar.addWidget(self.debugScriptButton)
        self.toolbar.addWidget(self.disasmScriptButton)
        self.toolbar.addAction(self.deadCodeScriptButton)
        self.toolbar.addWidget(undoSpacer)
        self.toolbar.addAction(self.__undoButton)
        self.toolbar.addAction(self.__redoButton)
        self.toolbar.addWidget(spacer)
        self.toolbar.addAction(self.removeTrailingSpacesButton)
        self.toolbar.addAction(self.expandTabsButton)

        self.importsBar = ImportListWidget(self.__editor)
        self.importsBar.hide()

        self.__outsideChangesBar = OutsideChangeWidget(self.__editor)
        self.__outsideChangesBar.sigReloadRequest.connect(self.__onReload)
        self.__outsideChangesBar.reloadAllNonModifiedRequest.connect(
            self.reloadAllNonModified)
        self.__outsideChangesBar.hide()

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 0, 0, 0)
        hLayout.setSpacing(0)

        vLayout = QVBoxLayout()
        vLayout.setContentsMargins(0, 0, 0, 0)
        vLayout.setSpacing(0)

        self.__navigationBar = NavigationBar(self.__editor, self)
        vLayout.addWidget(self.__navigationBar)
        vLayout.addWidget(self.__editor)

        hLayout.addLayout(vLayout)
        hLayout.addWidget(self.toolbar)
        widget = QWidget(self)
        widget.setLayout(hLayout)

        self.__splitter = QSplitter(Qt.Horizontal, self)

        self.__flowUI = FlowUIWidget(self.__editor, self)
        self.__mdView = MDWidget(self.__editor, self)

        self.__renderLayout = QVBoxLayout()
        self.__renderLayout.setContentsMargins(0, 0, 0, 0)
        self.__renderLayout.setSpacing(0)
        self.__renderLayout.addWidget(self.__flowUI)
        self.__renderLayout.addWidget(self.__mdView)
        self.__renderWidget = QWidget(self)
        self.__renderWidget.setLayout(self.__renderLayout)

        self.__splitter.addWidget(widget)
        self.__splitter.addWidget(self.__renderWidget)

        containerLayout = QHBoxLayout()
        containerLayout.setContentsMargins(0, 0, 0, 0)
        containerLayout.setSpacing(0)
        containerLayout.addWidget(self.__splitter)
        self.setLayout(containerLayout)

        self.__renderWidget.setVisible(False)

        self.__splitter.setSizes(Settings()['flowSplitterSizes'])
        self.__splitter.splitterMoved.connect(self.flowSplitterMoved)
        Settings().sigFlowSplitterChanged.connect(self.otherFlowSplitterMoved)

    def flowSplitterMoved(self, pos, index):
        """Splitter has been moved"""
        del pos     # unused argument
        del index   # unused argument
        Settings()['flowSplitterSizes'] = list(self.__splitter.sizes())

    def otherFlowSplitterMoved(self):
        """Other window has changed the splitter position"""
        self.__splitter.setSizes(Settings()['flowSplitterSizes'])

    def updateStatus(self):
        """Updates the toolbar buttons status"""
        self.__updateRunDebugButtons()
        isPythonFile = isPythonMime(self.__editor.mime)
        self.importsDiagramButton.setEnabled(
            isPythonFile and GlobalData().graphvizAvailable)
        self.__editor.diagramsMenu.setEnabled(
            self.importsDiagramButton.isEnabled())
        self.__editor.toolsMenu.setEnabled(self.runScriptButton.isEnabled())

    def onNavigationBar(self):
        """Triggered when navigation bar focus is requested"""
        if self.__navigationBar.isVisible():
            self.__navigationBar.setFocusToLastCombo()
        return True

    def __onPrint(self):
        """Triggered when the print button is pressed"""
        self.__editor._onShortcutPrint()

    def __onPrintPreview(self):
        """Triggered when the print preview button is pressed"""
        pass

    def __onDeadCode(self):
        """Triggered when vulture analysis is requested"""
        GlobalData().mainWindow.tabDeadCodeClicked()

    def __redoAvailable(self, available):
        """Reports redo ops available"""
        self.__redoButton.setEnabled(available)

    def __undoAvailable(self, available):
        """Reports undo ops available"""
        self.__undoButton.setEnabled(available)

    def __languageChanged(self, _=None):
        """Language changed"""
        isPython = self.__editor.isPythonBuffer()
        isMarkdown = self.__editor.isMarkdownBuffer()
        self.disasmScriptButton.setEnabled(isPython)
        self.__renderWidget.setVisible(not Settings()['floatingRenderer'] and
                                       (isPython or isMarkdown))

    # Arguments: modified
    def modificationChanged(self, _=None):
        """Triggered when the content is changed"""
        self.__updateRunDebugButtons()

    def __updateRunDebugButtons(self):
        """Enables/disables the run and debug buttons as required"""
        enable = isPythonMime(self.__editor.mime) and \
                 not self.isModified() and \
                 not self.__debugMode and \
                 os.path.isabs(self.__fileName)

        if enable != self.runScriptButton.isEnabled():
            self.runScriptButton.setEnabled(enable)
            self.profileScriptButton.setEnabled(enable)
            self.debugScriptButton.setEnabled(enable)
            self.deadCodeScriptButton.setEnabled(enable)
            self.sigTabRunChanged.emit(enable)

    def isTabRunEnabled(self):
        """Tells the status of run-like buttons"""
        return self.runScriptButton.isEnabled()

    def replaceAll(self, newText):
        """Replaces the current buffer content with a new text"""
        # Unfortunately, the setText() clears the undo history so it cannot be
        # used. The selectAll() and replacing selected text do not suite
        # because after undo the cursor does not jump to the previous position.
        # So, there is an ugly select -> replace manipulation below...
        with self.__editor:
            origLine, origPos = self.__editor.cursorPosition
            self.__editor.setSelection(0, 0, origLine, origPos)
            self.__editor.removeSelectedText()
            self.__editor.insert(newText)
            self.__editor.setCurrentPosition(len(newText))
            line, pos = self.__editor.cursorPosition
            lastLine = self.__editor.lines()
            self.__editor.setSelection(line, pos,
                                       lastLine - 1,
                                       len(self.__editor.text(lastLine - 1)))
            self.__editor.removeSelectedText()
            self.__editor.cursorPosition = origLine, origPos

            # These two for the proper cursor positioning after redo
            self.__editor.insert("s")
            self.__editor.cursorPosition = origLine, origPos + 1
            self.__editor.deleteBack()
            self.__editor.cursorPosition = origLine, origPos

    def onRemoveTrailingWS(self):
        """Triggers when the trailing spaces should be wiped out"""
        self.__editor.removeTrailingWhitespaces()

    def onExpandTabs(self):
        """Expands tabs if there are any"""
        self.__editor.expandTabs(4)

    def setFocus(self):
        """Overridden setFocus"""
        if self.__outsideChangesBar.isHidden():
            self.__editor.setFocus()
        else:
            self.__outsideChangesBar.setFocus()

    def onImportDgmTuned(self):
        """Runs the settings dialog first"""
        if self.isModified():
            what = ImportsDiagramDialog.SingleBuffer
            if not os.path.isabs(self.getFileName()):
                logging.warning("Imports diagram can only be generated for "
                                "a file. Save the editor buffer "
                                "and try again.")
                return
        else:
            what = ImportsDiagramDialog.SingleFile
        dlg = ImportsDiagramDialog(what, self.getFileName(), self)
        if dlg.exec_() == QDialog.Accepted:
            # Should proceed with the diagram generation
            self.__generateImportDiagram(what, dlg.options)

    # Arguments: action
    def onImportDgm(self, _=None):
        """Runs the generation process with default options"""
        if self.isModified():
            what = ImportsDiagramDialog.SingleBuffer
            if not os.path.isabs(self.getFileName()):
                logging.warning("Imports diagram can only be generated for "
                                "a file. Save the editor buffer "
                                "and try again.")
                return
        else:
            what = ImportsDiagramDialog.SingleFile
        self.__generateImportDiagram(what, ImportDiagramOptions())

    def __generateImportDiagram(self, what, options):
        """Show the generation progress and display the diagram"""
        if self.isModified():
            progressDlg = ImportsDiagramProgress(what, options,
                                                 self.getFileName(),
                                                 self.__editor.text)
            tooltip = "Generated for modified buffer (" + \
                      self.getFileName() + ")"
        else:
            progressDlg = ImportsDiagramProgress(what, options,
                                                 self.getFileName())
            tooltip = "Generated for file " + self.getFileName()
        if progressDlg.exec_() == QDialog.Accepted:
            GlobalData().mainWindow.openDiagram(progressDlg.scene,
                                                tooltip)

    def onOpenImport(self):
        """Triggered when Ctrl+I is received"""
        if isPythonMime(self.__editor.mime):
            # Take all the file imports and resolve them
            fileImports = getImportsList(self.__editor.text)
            if not fileImports:
                GlobalData().mainWindow.showStatusBarMessage(
                    "There are no imports")
            else:
                self.__onImportList(self.__fileName, fileImports)

    def __onImportList(self, fileName, imports):
        """Works with a list of imports"""
        # It has already been checked that the file is a Python one
        resolvedList, errors = resolveImports(fileName, imports)
        del errors  # errors are OK here
        if resolvedList:
            # Display the import selection widget
            self.importsBar.showResolvedImports(resolvedList)
        else:
            GlobalData().mainWindow.showStatusBarMessage(
                "Could not resolve any imports")

    def resizeEvent(self, event):
        """Resizes the import selection dialogue if necessary"""
        self.__editor.hideCompleter()
        QWidget.resizeEvent(self, event)
        self.resizeBars()

    def resizeBars(self):
        """Resize the bars if they are shown"""
        if not self.importsBar.isHidden():
            self.importsBar.resize()
        if not self.__outsideChangesBar.isHidden():
            self.__outsideChangesBar.resize()
        self.__editor.resizeCalltip()

    def showOutsideChangesBar(self, allEnabled):
        """Shows the bar for the editor for the user to choose the action"""
        self.setReloadDialogShown(True)
        self.__outsideChangesBar.showChoice(self.isModified(),
                                            allEnabled)

    def __onReload(self):
        """Triggered when a request to reload the file is received"""
        self.sigReloadRequest.emit()

    def reload(self):
        """Called (from the editors manager) to reload the file"""
        # Re-read the file with updating the file timestamp
        self.readFile(self.__fileName)

        # Hide the bars, just in case both of them
        if not self.importsBar.isHidden():
            self.importsBar.hide()
        if not self.__outsideChangesBar.isHidden():
            self.__outsideChangesBar.hide()

        # Set the shown flag
        self.setReloadDialogShown(False)

    def reloadAllNonModified(self):
        """Request to reload all the non-modified files"""
        self.reloadAllNonModifiedRequest.emit()

    @staticmethod
    def onRunScript(action=None):
        """Runs the script"""
        del action  # unused argument
        GlobalData().mainWindow.onRunTab()

    @staticmethod
    def onRunScriptDlg():
        """Shows the run parameters dialogue"""
        GlobalData().mainWindow.onRunTabDlg()

    @staticmethod
    def onProfileScript(action=None):
        """Profiles the script"""
        del action  # unused argument
        GlobalData().mainWindow.onProfileTab()

    @staticmethod
    def onProfileScriptDlg():
        """Shows the profile parameters dialogue"""
        GlobalData().mainWindow.onProfileTabDlg()

    @staticmethod
    def onDebugScript(action=None):
        """Starts debugging"""
        del action  # unused argument
        GlobalData().mainWindow.onDebugTab()

    @staticmethod
    def onDebugScriptDlg():
        """Shows the debug parameters dialogue"""
        GlobalData().mainWindow.onDebugTabDlg()

    def getCFEditor(self):
        """Provides a reference to the control flow widget"""
        return self.__flowUI

    def cflowSyncRequested(self, absPos, line, pos):
        """Highlight the item closest to the absPos"""
        self.__flowUI.highlightAtAbsPos(absPos, line, pos)

    def passFocusToFlow(self):
        """Sets the focus to the graphics part"""
        if isPythonMime(self.__editor.mime):
            self.__flowUI.setFocus()
            return True
        return False

    def getMDView(self):
        """Provides a reference to the MD rendered view"""
        return self.__mdView

    def terminate(self):
        """Called just before the tab is closed"""
        self.__splitter.splitterMoved.disconnect(self.flowSplitterMoved)
        Settings().sigFlowSplitterChanged.disconnect(
            self.otherFlowSplitterMoved)
        self.__cleanupLayout()

        self.__navigationBar.terminate()
        self.__navigationBar.deleteLater()

        self.__mdView.terminate()
        self.__mdView.deleteLater()

        self.__flowUI.terminate()
        self.__flowUI.deleteLater()

        self.__editor.terminate()
        self.__editor.deleteLater()

    def __cleanupLayout(self):
        """Disconnects QT widget signals and destroys the UI items"""
        self.__editor.redoAvailable.disconnect(self.__redoAvailable)
        self.__editor.undoAvailable.disconnect(self.__undoAvailable)
        self.__editor.modificationChanged.disconnect(self.modificationChanged)
        self.__editor.sigCFlowSyncRequested.disconnect(self.cflowSyncRequested)
        self.__editor.languageChanged.disconnect(self.__languageChanged)

        printButton = self.toolbar.findChild(QAction, 'printButton')
        printButton.triggered.disconnect(self.__onPrint)
        printButton.deleteLater()

        printPreviewButton = self.toolbar.findChild(QAction, 'printPreviewButton')
        printPreviewButton.triggered.connect(self.__onPrintPreview)
        printPreviewButton.deleteLater()

        self.importsDlgAct.triggered.disconnect(self.onImportDgmTuned)
        self.importsDlgAct.deleteLater()
        self.importsDiagramButton.menu().deleteLater()
        self.importsDiagramButton.clicked.disconnect(self.onImportDgm)
        self.importsDiagramButton.deleteLater()

        self.runScriptDlgAct.triggered.disconnect(self.onRunScriptDlg)
        self.runScriptDlgAct.deleteLater()
        self.runScriptButton.menu().deleteLater()
        self.runScriptButton.clicked.disconnect(self.onRunScript)
        self.runScriptButton.deleteLater()

        self.profileScriptDlgAct.triggered.disconnect(self.onProfileScriptDlg)
        self.profileScriptDlgAct.deleteLater()
        self.profileScriptButton.menu().deleteLater()
        self.profileScriptButton.clicked.disconnect(self.onProfileScript)
        self.profileScriptButton.deleteLater()

        self.debugScriptDlgAct.triggered.disconnect(self.onDebugScriptDlg)
        self.debugScriptDlgAct.deleteLater()
        self.debugScriptButton.menu().deleteLater()
        self.debugScriptButton.clicked.disconnect(self.onDebugScript)
        self.debugScriptButton.deleteLater()

        self.disasmScriptButton.clicked.disconnect(self.__editor._onDisasm0)
        self.disasmScriptButton.menu().deleteLater()
        self.disasmScriptButton.deleteLater()

        self.deadCodeScriptButton.triggered.disconnect(self.__onDeadCode)
        self.deadCodeScriptButton.deleteLater()

        self.__undoButton.triggered.disconnect(self.__editor.onUndo)
        self.__undoButton.deleteLater()

        self.__redoButton.triggered.disconnect(self.__editor.onRedo)
        self.__redoButton.deleteLater()

        self.removeTrailingSpacesButton.triggered.disconnect(
            self.onRemoveTrailingWS)
        self.removeTrailingSpacesButton.deleteLater()

        self.expandTabsButton.triggered.disconnect(self.onExpandTabs)
        self.expandTabsButton.deleteLater()

        self.__renderWidget.deleteLater()

        self.importsBar.deleteLater()
        self.toolbar.deleteLater()


    # Mandatory interface part is below

    def getEditor(self):
        """Provides the editor widget"""
        return self.__editor

    def isModified(self):
        """Tells if the file is modified"""
        return self.__editor.document().isModified()

    def getRWMode(self):
        """Tells if the file is read only"""
        if not os.path.exists(self.__fileName):
            return None
        return 'RW' if QFileInfo(self.__fileName).isWritable() else 'RO'

    def getMime(self):
        """Provides the buffer mime"""
        return self.__editor.mime

    @staticmethod
    def getType():
        """Tells the widget type"""
        return MainWindowTabWidgetBase.PlainTextEditor

    def getLanguage(self):
        """Tells the content language"""
        editorLanguage = self.__editor.language()
        if editorLanguage:
            return editorLanguage
        return self.__editor.mime if self.__editor.mime else 'n/a'

    def getFileName(self):
        """Tells what file name of the widget content"""
        return self.__fileName

    def setFileName(self, name):
        """Sets the file name"""
        self.__fileName = name
        self.__shortName = os.path.basename(name)

    def getEol(self):
        """Tells the EOL style"""
        return self.__editor.getEolIndicator()

    def getLine(self):
        """Tells the cursor line"""
        line, _ = self.__editor.cursorPosition
        return line

    def getPos(self):
        """Tells the cursor column"""
        _, pos = self.__editor.cursorPosition
        return pos

    def getEncoding(self):
        """Tells the content encoding"""
        if self.__editor.explicitUserEncoding:
            return self.__editor.explicitUserEncoding
        return self.__editor.encoding

    def getShortName(self):
        """Tells the display name"""
        return self.__shortName

    def setShortName(self, name):
        """Sets the display name"""
        self.__shortName = name

    def isDiskFileModified(self):
        """Return True if the loaded file is modified"""
        if not os.path.isabs(self.__fileName):
            return False
        if not os.path.exists(self.__fileName):
            return True
        path = os.path.realpath(self.__fileName)
        return self.__diskModTime != os.path.getmtime(path) or \
               self.__diskSize != os.path.getsize(path)

    def doesFileExist(self):
        """Returns True if the loaded file still exists"""
        return os.path.exists(self.__fileName)

    def setReloadDialogShown(self, value=True):
        """Memorizes if the reloading dialogue has already been displayed"""
        self.__reloadDlgShown = value

    def getReloadDialogShown(self):
        """Tells if the reload dialog has already been shown"""
        return self.__reloadDlgShown and \
            not self.__outsideChangesBar.isVisible()

    def updateModificationTime(self, fileName):
        """Updates the modification time"""
        path = os.path.realpath(fileName)
        self.__diskModTime = os.path.getmtime(path)
        self.__diskSize = os.path.getsize(path)

    def setDebugMode(self, debugOn, disableEditing):
        """Called to switch debug/development"""
        self.__debugMode = debugOn
        self.__editor.setDebugMode(debugOn, disableEditing)

        if debugOn:
            if disableEditing:
                # Undo/redo
                self.__undoButton.setEnabled(False)
                self.__redoButton.setEnabled(False)

                # Spaces/tabs/line
                self.removeTrailingSpacesButton.setEnabled(False)
                self.expandTabsButton.setEnabled(False)
        else:
            # Undo/redo
            self.__undoButton.setEnabled(
                self.__editor.document().isUndoAvailable())
            self.__redoButton.setEnabled(
                self.__editor.document().isRedoAvailable())

            # Spaces/tabs
            self.removeTrailingSpacesButton.setEnabled(True)
            self.expandTabsButton.setEnabled(True)

        # Run/debug buttons
        self.__updateRunDebugButtons()

    def isLineBreakable(self, line=None, enforceRecalc=False,
                        enforceSure=False):
        """True if a breakpoint could be placed on the current line"""
        return self.__editor.isLineBreakable()

    def getVCSStatus(self):
        """Provides the VCS status"""
        return self.__vcsStatus

    def setVCSStatus(self, newStatus):
        """Sets the new VCS status"""
        self.__vcsStatus = newStatus

    # Floating renderer support
    def popRenderingWidgets(self):
        """Pops the rendering widgets"""
        self.__renderLayout.removeWidget(self.__flowUI)
        self.__renderLayout.removeWidget(self.__mdView)
        self.__renderWidget.setVisible(False)
        return [self.__flowUI, self.__mdView]

    def pushRenderingWidgets(self, widgets):
        """Returns back the rendering widgets"""
        for widget in widgets:
            self.__renderLayout.addWidget(widget)
        self.__languageChanged()    # Sets the widget visibility



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2019 Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os.path
from ui.qt import (Qt, QDialog, QDialogButtonBox, QVBoxLayout, QLabel,
                   QPushButton, QGridLayout, QLineEdit, QTextEdit, QCheckBox,
                   QFileDialog)
from utils.colorfont import setLineEditBackground, getZoomedMonoFont
from utils.globals import GlobalData
from utils.misc import preResolveLinkPath


"""Dialog to enter a new text for a graphics item"""

class DocLinkAnchorDialog(QDialog):

    """Replace text input dialog"""

    def __init__(self, windowTitle, cmlDocComment, fileName, parent):
        QDialog.__init__(self, parent)

        # Name of a file from which the doc link is created/edited
        self.__fileName = fileName
        self.setWindowTitle(windowTitle + ' documentation link and/or anchor')

        self.__createLayout()
        self.__invalidInputColor = GlobalData().skin['invalidInputPaper']
        self.__validInputColor = self.linkEdit.palette().color(
            self.linkEdit.backgroundRole())

        if cmlDocComment is not None:
            self.__populate(cmlDocComment)

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(450, 150)
        self.setSizeGripEnabled(True)

        verticalLayout = QVBoxLayout(self)
        gridLayout = QGridLayout()

        # Link
        gridLayout.addWidget(QLabel('Link', self), 0, 0, 1, 1)
        self.linkEdit = QLineEdit(self)
        self.linkEdit.setClearButtonEnabled(True)
        self.linkEdit.setToolTip(
            'A link to a file or to an external web resource')
        gridLayout.addWidget(self.linkEdit, 0, 1, 1, 1)
        self.linkEdit.textChanged.connect(self.__validate)
        self.fileButton = QPushButton(self)
        self.fileButton.setText('...')
        self.fileButton.setToolTip(
            'Select an existing or non existing file')
        gridLayout.addWidget(self.fileButton, 0, 2, 1, 1)
        self.fileButton.clicked.connect(self.__onSelectPath)
        self.createCheckBox = QCheckBox(
            'Create a markdown file if does not exist', self)
        self.createCheckBox.setChecked(False)
        gridLayout.addWidget(self.createCheckBox, 1, 1, 1, 1)
        self.createCheckBox.stateChanged.connect(self.__validate)

        # Anchor
        gridLayout.addWidget(QLabel('Anchor', self), 2, 0, 1, 1)
        self.anchorEdit = QLineEdit(self)
        self.anchorEdit.setClearButtonEnabled(True)
        gridLayout.addWidget(self.anchorEdit, 2, 1, 1, 1)
        self.anchorEdit.textChanged.connect(self.__validate)

        # Title
        titleLabel = QLabel('Title', self)
        titleLabel.setAlignment(Qt.AlignTop)
        gridLayout.addWidget(titleLabel, 3, 0, 1, 1)
        self.titleEdit = QTextEdit()
        self.titleEdit.setTabChangesFocus(True)
        self.titleEdit.setAcceptRichText(False)
        self.titleEdit.setFont(getZoomedMonoFont())
        self.titleEdit.setToolTip(
            'If provided then will be displayed in the rectangle')
        gridLayout.addWidget(self.titleEdit, 3, 1, 1, 1)

        # Buttons at the bottom
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Ok)
        self.__OKButton = buttonBox.button(QDialogButtonBox.Ok)
        self.__OKButton.setDefault(True)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.close)

        verticalLayout.addLayout(gridLayout)
        verticalLayout.addWidget(buttonBox)

        self.linkEdit.setFocus()

    def setTitle(self, txt):
        """Sets the title text to be edited"""
        self.titleEdit.setPlainText(txt)

    def title(self):
        """Provides the new title text"""
        return self.titleEdit.toPlainText()

    def needToCreate(self):
        return self.createCheckBox.isEnabled() and self.createCheckBox.isChecked()

    def __populate(self, cmlDocComment):
        """Populates the fields from the comment"""
        if cmlDocComment.link:
            self.linkEdit.setText(cmlDocComment.link)
        if cmlDocComment.anchor:
            self.anchorEdit.setText(cmlDocComment.anchor)
        if cmlDocComment.title:
            self.setTitle(cmlDocComment.getTitle())

    def __onSelectPath(self):
        """Select file or directory"""
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        selectedPath = QFileDialog.getOpenFileName(
            self, 'Select documentation file', self.linkEdit.text(),
            options=options)
        if isinstance(selectedPath, tuple):
            selectedPath = selectedPath[0]
        if selectedPath:
            self.linkEdit.setText(os.path.normpath(selectedPath))

    def __setLinkValid(self):
        """Sets the link edit valid"""
        self.linkEdit.setToolTip(
            'A link to a file or to an external web resource')
        setLineEditBackground(self.linkEdit, self.__validInputColor,
                              self.__validInputColor)

    def __setLinkInvalid(self, msg):
        """Sets the link edit invalid"""
        self.linkEdit.setToolTip(msg)
        setLineEditBackground(self.linkEdit, self.__invalidInputColor,
                              self.__invalidInputColor)

    def __validateLink(self):
        """Validates the link field content"""
        txt = self.linkEdit.text().strip()
        if txt == '' or txt.startswith('http://') or txt.startswith('https://'):
            self.__setLinkValid()
            self.createCheckBox.setEnabled(False)
            return True

        if txt.endswith(os.path.sep):
            self.__setLinkInvalid('A link must be a file, not a directory')
            return False

        # Not a link; it is supposed to be a file or a creatable file
        # However the invalid values will also be acceptable
        self.createCheckBox.setEnabled(True)
        fromFile = None
        if self.__fileName:
            if os.path.isabs(self.__fileName):
                fromFile = self.__fileName
        fName, anchor, errMsg = preResolveLinkPath(
            txt, fromFile, self.createCheckBox.isChecked())
        del anchor

        if fName:
            self.__setLinkValid()
            return True

        self.__setLinkInvalid(errMsg)
        return not self.createCheckBox.isChecked()

    def __setAnchorValid(self):
        """Sets the anchor edit valid"""
        self.anchorEdit.setToolTip(
            'Anchor may not contain neither spaces nor tabs')
        setLineEditBackground(self.anchorEdit, self.__invalidInputColor,
                              self.__validInputColor)

    def __setAnchorInvalid(self):
        """Sets the anchor edit invalid"""
        self.anchorEdit.setToolTip(
            'Anchor is used to refer to it from the other files')
        setLineEditBackground(self.anchorEdit, self.__validInputColor,
                              self.__validInputColor)

    def __validateAnchor(self):
        """Validates the anchor field"""
        txt = self.anchorEdit.text().strip()
        if ' ' in txt or '\t' in txt:
            self.__setAnchorValid()
            return False
        self.__setAnchorInvalid()
        return True

    def __validate(self, _=None):
        """Validates the input fields and sets the OK button enable"""
        self.__OKButton.setToolTip('')
        valid = self.__validateAnchor() and self.__validateLink()
        if valid:
            if not self.linkEdit.text().strip() and not self.anchorEdit.text().strip():
                valid = False
                self.__OKButton.setToolTip(
                    'At least one of the items: link or anchor must be provided')

        self.__OKButton.setEnabled(valid)
        return valid




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Debugger breakpoint"""

import os, os.path
from utils.globals import GlobalData


class Breakpoint:

    """Represents a single breakpoint"""

    def __init__(self, fileName=None, lineNumber=None, condition="",
                 temporary=False, enabled=True, ignoreCount=0):
        if fileName is None:
            self.__fileName = fileName
        elif os.path.isabs(fileName):
            project = GlobalData().project
            if project.isLoaded():
                if project.isProjectFile(fileName):
                    # This is a project file; strip the project dir
                    self.__fileName = fileName.replace(
                        project.getProjectDir(), "")
                else:
                    # Not a project file, save as is
                    self.__fileName = fileName
            else:
                # Pretty much impossible
                self.__fileName = fileName
        else:
            # Relative path, i.e. a project file
            self.__fileName = fileName

        self.__lineNumber = lineNumber
        self.__condition = condition
        self.__temporary = temporary
        self.__enabled = enabled
        self.__ignoreCount = ignoreCount

    def update(self, otherBreakPoint):
        """Basically copies values from another breakpoint"""
        self.__fileName = otherBreakPoint.getFileName()
        self.__lineNumber = otherBreakPoint.getLineNumber()
        self.__condition = otherBreakPoint.getCondition()
        self.__temporary = otherBreakPoint.isTemporary()
        self.__enabled = otherBreakPoint.isEnabled()
        self.__ignoreCount = otherBreakPoint.getIgnoreCount()

    def updateLineNumber(self, line):
        """Updates the line #. Used when there are changes in the editor"""
        self.__lineNumber = line

    def isValid(self):
        """True if the breakpoint is valid"""
        if self.__fileName is None:
            return False

        if os.path.isabs(self.__fileName):
            if not os.path.exists(self.__fileName):
                return False
        else:
            project = GlobalData().project
            if project.isLoaded():
                path = project.getProjectDir() + self.__fileName
                if not os.path.exists(path):
                    return False
            else:
                if not os.path.exists(self.__fileName):
                    return False

        return self.__lineNumber is not None and self.__lineNumber > 0

    def getFileName(self):
        """Provides the file name"""
        return self.__fileName

    def getAbsoluteFileName(self):
        """Provides the absolute file name"""
        if self.__fileName is None:
            return None
        if os.path.isabs(self.__fileName):
            return self.__fileName

        project = GlobalData().project
        if project.isLoaded():
            return project.getProjectDir() + self.__fileName
        return os.path.abspath(self.__fileName)

    def getLineNumber(self):
        """Provides the line number"""
        return self.__lineNumber

    def getCondition(self):
        """Provides the condition"""
        return self.__condition

    def isTemporary(self):
        """True if temporary"""
        return self.__temporary

    def setTemporary(self, temp):
        """Sets the new value"""
        self.__temporary = temp

    def isEnabled(self):
        """True if enabled"""
        return self.__enabled

    def setEnabled(self, enabled):
        """Sets the new value"""
        self.__enabled = enabled

    def getIgnoreCount(self):
        """Provides the ignore count"""
        return self.__ignoreCount

    def getLocation(self, fullForm=False):
        """Provides the breakpoint location"""
        if self.__fileName is None:
            return str(self.__fileName) + ":" + str(self.__lineNumber)

        if fullForm:
            return self.getAbsoluteFileName() + ":" + str(self.__lineNumber)

        return os.path.basename(self.__fileName) + ":" + str(self.__lineNumber)

    def getTooltip(self):
        """Provides the breakpoint tooltip"""
        return "Location: " + self.getLocation(True) + "\n" \
               "Enabled: " + str(self.__enabled) + "\n" \
               "Temporary: " + str(self.__temporary) + "\n" \
               "Ignore count: " + str(self.__ignoreCount) + "\n" \
               "Condition: " + self.__condition

    def serialize(self):
        """Serializes the breakpoint to a string"""
        return {'file': self.__fileName,
                'line': self.__lineNumber,
                'condition': self.__condition,
                'temp': self.__temporary,
                'enabled': self.__enabled,
                'ignorecnt': self.__ignoreCount}

    def deserialize(self, source):
        """Deserializes the breakpoint"""
        self.__fileName = source.get('file', None)
        self.__lineNumber = source.get('line', None)
        self.__condition = source.get('condition', None)
        self.__temporary = source.get('temp', False)
        self.__enabled = source.get('enabled', False)
        self.__ignoreCount = source.get('ignorecnt', 0)
        return self.isValid()

    def __eq__(self, other):
        """True if equal"""
        return self.__fileName == other.__fileName and \
               self.__lineNumber == other.__lineNumber and \
               self.__condition == other.__condition and \
               self.__temporary == other.__temporary and \
               self.__enabled == other.__enabled and \
               self.__ignoreCount == other.__ignoreCount



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Control flow UI navigation bar"""

from ui.qt import Qt, QHBoxLayout, QLabel, QFrame, QSizePolicy
from ui.labels import HeaderFitLabel, HeaderLabel
from ui.spacers import ToolBarExpandingSpacer
from utils.pixmapcache import getPixmap


class ControlFlowNavigationBar(QFrame):

    """Navigation bar at the top of the flow UI widget"""

    STATE_OK_UTD = 0        # Parsed OK, control flow up to date
    STATE_OK_CHN = 1        # Parsed OK, control flow changed
    STATE_BROKEN_UTD = 2    # Parsed with errors, control flow up to date
    STATE_BROKEN_CHN = 3    # Parsed with errors, control flow changed
    STATE_UNKNOWN = 4

    def __init__(self, parent):
        QFrame.__init__(self, parent)
        self.__infoIcon = None
        self.__warningsIcon = None
        self.__layout = None
        self.__pathLabel = None
        self.__createLayout()
        self.__currentIconState = self.STATE_UNKNOWN

    def __createLayout(self):
        """Creates the layout"""
        self.__layout = QHBoxLayout(self)
        self.__layout.setContentsMargins(0, 0, 0, 0)

        # Create info icon
        self.__infoIcon = QLabel(self)
        self.__infoIcon.setPixmap(getPixmap('cfunknown.png'))
        self.__layout.addWidget(self.__infoIcon)

        self.__warningsIcon = QLabel(self)
        self.__warningsIcon.setPixmap(getPixmap('cfwarning.png'))
        self.__layout.addWidget(self.__warningsIcon)

        self.clearWarnings()

        # Create the path label
        self.__pathLabel = HeaderFitLabel(self)
        self.__pathLabel.setTextFormat(Qt.PlainText)
        self.__pathLabel.setAlignment(Qt.AlignLeft)
        self.__pathLabel.setWordWrap(False)
        self.__pathLabel.setTextInteractionFlags(Qt.NoTextInteraction)
        self.__pathLabel.setSizePolicy(QSizePolicy.Expanding,
                                       QSizePolicy.Fixed)
        self.__pathLabel.setMinimumWidth(40)
        self.__layout.addWidget(self.__pathLabel)

        self.__spacer = ToolBarExpandingSpacer(self)
        self.__spacer.setMinimumWidth(0)
        self.__layout.addWidget(self.__spacer)

        # Create the selection label
        self.__selectionLabel = HeaderLabel('', None, self)
        self.__selectionLabel.setTextFormat(Qt.PlainText)
        self.__selectionLabel.setAlignment(Qt.AlignCenter)
        self.__selectionLabel.setWordWrap(False)
        self.__selectionLabel.setTextInteractionFlags(Qt.NoTextInteraction)
        self.__selectionLabel.setSizePolicy(QSizePolicy.Fixed,
                                            QSizePolicy.Fixed)
        self.__selectionLabel.setMinimumWidth(40)
        self.__layout.addWidget(self.__selectionLabel)
        self.setSelectionLabel(0, None)

    def clearWarnings(self):
        """Clears the warnings"""
        self.__warningsIcon.setVisible(False)
        self.__warningsIcon.setToolTip("")

    def setWarnings(self, warnings):
        """Sets the warnings"""
        self.__warningsIcon.setToolTip(
            'Control flow parser warnings:\n' + '\n'.join(warnings))
        self.__warningsIcon.setVisible(True)

    def clearErrors(self):
        """Clears all the errors"""
        self.__infoIcon.setToolTip('')

    def setErrors(self, errors):
        """Sets the errors"""
        self.__infoIcon.setToolTip(
            'Control flow parser errors:\n' + '\n'.join(errors))

    def updateInfoIcon(self, state):
        """Updates the information icon"""
        if state == self.__currentIconState:
            return

        if state == self.STATE_OK_UTD:
            self.__infoIcon.setPixmap(getPixmap('cfokutd.png'))
            self.__infoIcon.setToolTip("Control flow is up to date")
            self.__currentIconState = self.STATE_OK_UTD
        elif state == self.STATE_OK_CHN:
            self.__infoIcon.setPixmap(getPixmap('cfokchn.png'))
            self.__infoIcon.setToolTip("Control flow is not up to date; "
                                       "will be updated on idle")
            self.__currentIconState = self.STATE_OK_CHN
        elif state == self.STATE_BROKEN_UTD:
            self.__infoIcon.setPixmap(getPixmap('cfbrokenutd.png'))
            self.__infoIcon.setToolTip("Control flow might be invalid "
                                       "due to invalid python code")
            self.__currentIconState = self.STATE_BROKEN_UTD
        elif state == self.STATE_BROKEN_CHN:
            self.__infoIcon.setPixmap(getPixmap('cfbrokenchn.png'))
            self.__infoIcon.setToolTip("Control flow might be invalid; "
                                       "will be updated on idle")
            self.__currentIconState = self.STATE_BROKEN_CHN
        else:
            # STATE_UNKNOWN
            self.__infoIcon.setPixmap(getPixmap('cfunknown.png'))
            self.__infoIcon.setToolTip("Control flow state is unknown")
            self.__currentIconState = self.STATE_UNKNOWN

    def getCurrentState(self):
        """Provides the current state"""
        return self.__currentIconState

    def setPath(self, txt):
        """Sets the path label content"""
        self.__pathLabel.setText(txt)

    def setPathVisible(self, switchOn):
        """Sets the path visible"""
        self.__pathLabel.setVisible(switchOn)
        self.__spacer.setVisible(not switchOn)

    def setSelectionLabel(self, text, tooltip):
        """Sets selection label"""
        self.__selectionLabel.setText(str(text))
        if tooltip:
            self.__selectionLabel.setToolTip("Selected items:\n" +
                                             str(tooltip))
        else:
            self.__selectionLabel.setToolTip("Number of selected items")

    def resizeEvent(self, event):
        """Editor has resized"""
        QFrame.resizeEvent(self, event)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""completer selection widget"""


from .qt import (Qt, QEvent, QModelIndex, QStringListModel, QAction,
                 QAbstractItemView, QCompleter, QListView)
from .itemdelegates import NoOutlineHeightDelegate


class CompleterPopup(QListView):

    """Custom completer popup"""

    def __init__(self, completer):
        QListView.__init__(self, completer.parent())

        self.__completer = completer
        self.setAlternatingRowColors(True)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        shiftTabAct = QAction(self)
        shiftTabAct.setShortcut('Shift+Tab')
        shiftTabAct.triggered.connect(self.__completer.moveToPrevious)
        self.addAction(shiftTabAct)

        ctrlSpaceAct = QAction(self)
        ctrlSpaceAct.setShortcut('Ctrl+ ')
        ctrlSpaceAct.triggered.connect(self.__completer.moveToNext)
        self.addAction(ctrlSpaceAct)


class WordsListModel(QStringListModel):

    """Custom model to use the same font as the editor"""

    def __init__(self, words, font):
        words.sort()
        QStringListModel.__init__(self, words)
        self.__font = font
        self.__font.setBold(False)

    def data(self, index, role):
        """Changes the rows font"""
        if role == Qt.FontRole:
            return self.__font
        return QStringListModel.data(self, index, role)


class CodeCompleter(QCompleter):

    """Codimension code completer"""

    maxDisplayedItems = 8

    def __init__(self, parent=None):
        QCompleter.__init__(self, parent)
        self.setCompletionMode(QCompleter.PopupCompletion)

        self.__width = -1
        self.__rect = None
        self.__model = None

        self.setWidget(parent)
        self.setCaseSensitivity(Qt.CaseInsensitive)
        self.setWrapAround(False)

        self.highlighted[QModelIndex].connect(self.__onHighlighted)

    def __onHighlighted(self, index):
        """Sets the current row to have currentCompletion() working properly"""
        self.setCurrentRow(index.row())

    def setWordsList(self, words, font):
        """Sets the words list"""
        self.setPopup(CompleterPopup(self))
        self.popup().setAlternatingRowColors(True)

        self.__model = WordsListModel(words, font)
        self.setModel(self.__model)
        self.setCompletionPrefix("")

        # This call should be given here, i.e. after a new model is set for the
        # completer. Otherwise this will have no effect.
        self.popup().setItemDelegate(NoOutlineHeightDelegate(4))

        self.__width = self.popup().sizeHintForColumn(0) + \
            self.popup().verticalScrollBar().sizeHint().width() + 20
        if self.__width < 100:
            self.__width = 100

    def setPrefix(self, prefix):
        """Sets the prefix"""
        self.setCompletionPrefix(str(prefix))

        count = self.completionCount()
        if count == 0:
            return

        # Make the completer resizable as the number of options changed
        if count < CodeCompleter.maxDisplayedItems:
            if self.maxVisibleItems() != count:
                self.setMaxVisibleItems(count)
                if self.popup().isVisible():
                    self.complete(self.__rect)
        else:
            if self.maxVisibleItems() != CodeCompleter.maxDisplayedItems:
                self.setMaxVisibleItems(CodeCompleter.maxDisplayedItems)
                if self.popup().isVisible():
                    self.complete(self.__rect)

        if self.completionCount() > 0:
            self.__selectFirst()

        self.parent().setFocus()

    def complete(self, rect):
        """Brings the completer up"""
        rect.setWidth(self.__width)
        self.__rect = rect
        QCompleter.complete(self, rect)
        self.__selectFirst()

    def __selectFirst(self):
        """Selects the first item in the list view if nothing is selected"""
        if self.completionCount() <= 0:
            return
        index = self.popup().currentIndex()
        if not index.isValid():
            cidx = self.completionModel().index(0, 0)
            self.popup().setCurrentIndex(cidx)

    def moveToNext(self):
        """Moves the selection one row down"""
        if self.completionCount() <= 1:
            return

        newRow = self.popup().currentIndex().row() + 1
        lastRow = self.completionCount() - 1
        if newRow <= lastRow:
            cidx = self.completionModel().index(newRow, 0)
        else:
            cidx = self.completionModel().index(0, 0)
        self.popup().setCurrentIndex(cidx)

    def moveToPrevious(self):
        """Moves the selection one row up"""
        if self.completionCount() <= 1:
            return

        newRow = self.popup().currentIndex().row() - 1
        if newRow >= 0:
            cidx = self.completionModel().index(newRow, 0)
        else:
            lastRow = self.completionCount() - 1
            cidx = self.completionModel().index(lastRow, 0)
        self.popup().setCurrentIndex(cidx)

    def eventFilter(self, obj, evnt):
        """Custom events filtering"""
        if evnt.type() == QEvent.KeyPress and self.popup().isVisible():
            if evnt.modifiers() != Qt.NoModifier and len(evnt.text()) == 0:
                # Supress
                return True
            key = evnt.key()
            if key in [Qt.Key_Enter, Qt.Key_Return]:
                self.activated.emit(self.currentCompletion())
                return True
            elif key == Qt.Key_Home:
                cidx = self.completionModel().index(0, 0)
                self.popup().setCurrentIndex(cidx)
                return True
            elif key == Qt.Key_End:
                lastRow = self.completionCount() - 1
                cidx = self.completionModel().index(lastRow, 0)
                self.popup().setCurrentIndex(cidx)
                return True
            elif key in [Qt.Key_Tab, Qt.Key_Down, Qt.Key_Right]:
                self.moveToNext()
                return True
            elif key in [Qt.Key_Backtab, Qt.Key_Up, Qt.Key_Left]:
                self.moveToPrevious()
                return True
            elif key == Qt.Key_PageUp:
                cRow = self.popup().currentIndex().row()
                if cRow == 0:
                    self.moveToPrevious()
                    return True
            elif key == Qt.Key_PageDown:
                lastRow = self.completionCount() - 1
                cRow = self.popup().currentIndex().row()
                if lastRow == cRow:
                    self.moveToNext()
                    return True

        return QCompleter.eventFilter(self, obj, evnt)

    def isVisible(self):
        """Returns True if the popup is visible"""
        return self.popup().isVisible()

    def hide(self):
        """Hides the popup"""
        self.popup().hide()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Miscellaneuos utility functions"""

# pylint: disable=W0702
# pylint: disable=W0703

import os.path
import re
import getpass
import locale
import datetime
import logging
import traceback
import gc
from .globals import GlobalData
from .settings import SETTINGS_DIR
from .fileutils import getFileContent, isFileOpenable, isCreatable


# File name of the template for any new file.
# The file is searched nearby the project file.
templateFileName = 'template.py'

# File name with pylint settings
pylintFileName = 'pylintrc'


def splitThousands(value, sep="'"):
    """Provides thousands separated value"""
    if len(value) <= 3:
        return value
    return splitThousands(value[:-3], sep) + sep + value[-3:]


def getLocaleDate():
    """Provides locale formatted date"""
    now = datetime.datetime.now()
    try:
        date_format = locale.nl_langinfo(locale.D_FMT)
        return now.strftime(date_format)
    except:
        return now.strftime('%Y-%m-%d')


def getLocaleTime():
    """Provides locale formatted time"""
    now = datetime.datetime.now()
    try:
        time_format = locale.nl_langinfo(locale.T_FMT)
        return now.strftime(time_format)
    except:
        return now.strftime('%H:%M:%S')


def getLocaleDateTime():
    """Provides locale date time"""
    return getLocaleDate() + " " + getLocaleTime()


def getIDETemplateFile():
    """Provides the name of the IDE template file"""
    return SETTINGS_DIR + templateFileName


def getIDEPylintFile():
    """Provides the name of the IDE pylintrc file"""
    return SETTINGS_DIR + pylintFileName


def getProjectTemplateFile():
    """Provides the name of the project template file"""
    project = GlobalData().project
    if project.isLoaded():
        # Project is loaded - use from the project dir
        projectDir = os.path.dirname(project.fileName)
        if not projectDir.endswith(os.path.sep):
            projectDir += os.path.sep
        return projectDir + templateFileName
    return None


def getNewFileTemplate():
    """Searches for the template file and fills fields in it"""
    templateFile = getProjectTemplateFile()
    if templateFile is None:
        templateFile = getIDETemplateFile()
    elif not os.path.exists(templateFile):
        templateFile = getIDETemplateFile()

    if not os.path.exists(templateFile):
        return ""

    # read the file content: splitlines() eats the trailing \n
    content = getFileContent(templateFile)
    if content.endswith('\n') or content.endswith('\r\n'):
        content = content.splitlines() + ['']
    else:
        content = content.splitlines()

    # substitute the fields
    project = GlobalData().project
    projectLoaded = project.isLoaded()
    if projectLoaded:
        subs = [(re.compile(re.escape('$projectdate'), re.I),
                 project.props['creationdate']),
                (re.compile(re.escape('$author'), re.I),
                 project.props['author']),
                (re.compile(re.escape('$license'), re.I),
                 project.props['license']),
                (re.compile(re.escape('$copyright'), re.I),
                 project.props['copyright']),
                (re.compile(re.escape('$version'), re.I),
                 project.props['version']),
                (re.compile(re.escape('$email'), re.I),
                 project.props['email'])]
    else:
        subs = []

    # Common substitutions
    subs.append((re.compile(re.escape('$date'), re.I),
                 getLocaleDate()))
    subs.append((re.compile(re.escape('$time'), re.I),
                 getLocaleTime()))
    subs.append((re.compile(re.escape('$user'), re.I),
                 getpass.getuser()))

    if projectLoaded:
        # description could be multilined so it is a different story
        descriptionRegexp = re.compile(re.escape('$description'), re.I)
        description = project.props['description'].split('\n')

    result = []
    for line in content:
        for key, value in subs:
            line = re.sub(key, value, line)
        if projectLoaded:
            # description part if so
            match = re.search(descriptionRegexp, line)
            if match is not None:
                # description is in the line
                leadingPart = line[:match.start()]
                trailingPart = line[match.end():]
                for dline in description:
                    result.append(leadingPart + dline + trailingPart)
            else:
                result.append(line)
        else:
            result.append(line)

    return '\n'.join(result)


def getDefaultTemplate():
    """Provides a body (i.e. help) of the default template file"""
    return """#
# This template will be used when a new file is created.
#
# Codimension supports an IDE-wide template file and project-specific template
# files for each project. If a project is loaded then codimension checks the
# project-specific template file and IDE-wide one in this order. The first one
# found is used to create a new file. If no project is loaded then only the
# IDE-wide template file is checked.
# The IDE-wide template file is stored in the codimension settings directory
# while project-specific template files are stored in the top project
# directory. In both cases the template file is called 'template.py'.
#
# The following variables will be replaced with actual values if
# they are found in the template:
#
# $projectdate     Project creation date     (*)
# $author          Project author            (*)
# $license         Project license           (*)
# $copyright       Project copyright string  (*)
# $version         Project version           (*)
# $email           Project author e-mail     (*)
# $description     Project description       (*)
# $date            Current date
# $time            Current time
# $user            Current user name
#
# Note: variables marked with (*) are available only for the project-specific
#       templates. The values for the variables are taken from the project
#       properties dialogue.
#"""


def getDefaultProjectDoc(fName):
    """Provides a body (i.e. help) of the default project doc"""
    return """The project documentation is not found.

Codimension supports the markdown format for documentation purposes
and there are a few options to create and specify the project documentation
start point:

- create an .md file anywhere on the file system and specify the path to it in
  the project properties
- simply create the README.md in the project root directory (%s)

Please discard these instructions, provide the required content and save as
needed. If you do so, next time the project doc button is clicked, the project
doc markdown file will be displayed.
""".replace('%s', fName)


def getDefaultFileDoc(fName, anchor):
    """Provides a body (i.e. help) of the default auto doc"""
    return """Codimension supports markdown for the documentation purposes.

A doc link to this file has been inserted into your python file. When it
is clicked on the graphics pane this file will be displayed.

The source code doc link also has a generated anchor (${anchor}) so the
documentation can refer specifically to that doc link. To do so use the
following format:

```
[See in the source code](file:${fName}#${anchor})
```

When clicked, the source code file will be opened and scrolled to the doc link.

Please discard these instructions, provide the required content and save as
needed.
""".replace('${fName}', fName).replace('${anchor}', anchor)


# Dynamic mixin at runtime:
# http://stackoverflow.com/questions/8544983/
#        dynamically-mixin-a-base-class-to-an-instance-in-python
def extendInstance(obj, cls):
    """dynamic mixin support"""
    base_cls = obj.__class__
    base_cls_name = obj.__class__.__name__
    obj.__class__ = type(base_cls_name, (base_cls, cls), {})


ANCHOR_REGEXP = re.compile(r'\#[_a-zA-Z0-9]+$')

# Supported format:
# [file:]<absolute or relative path>[#anchor identifier]
def splitLinkPath(link):
    """Splits the link path into the path and the anchor"""
    if link.startswith('file:'):
        link = link[5:]
    link = os.path.normpath(link)

    anchor = None
    match = ANCHOR_REGEXP.search(link)
    if match is not None:
        anchorPart = match.group()
        anchor = anchorPart[1:]
        link = link[0:-1 * len(anchorPart)].strip()
    return link, anchor


def resolveFile(link, fromFile):
    """Tries to reslve file

    Returns: file name, list of tried paths
    """
    if os.path.isabs(link):
        if os.path.exists(link):
            return link, [link]
        return None, [link]

    tryPaths = []
    if fromFile:
        # Try relative to the 'fromFile' first
        dirName = os.path.dirname(fromFile)
        fName = os.path.normpath(dirName + os.path.sep + link)
        if os.path.exists(fName):
            return fName, [fName]
        tryPaths.append(fName)

    # Try relative to the project file second
    project = GlobalData().project
    if project.isLoaded():
        # Project is loaded - use from the project dir
        projectDir = os.path.dirname(project.fileName)
        fName = os.path.normpath(projectDir + os.path.sep + link)
        if fName not in tryPaths:
            if os.path.exists(fName):
                return fName, [fName]
            tryPaths.append(fName)

    return None, tryPaths


def checkExistingLinkTarget(fName):
    """Returns an error message or None if everything is fine"""
    if not os.path.isfile(fName):
        return ' '.join(("The resolved link path",
                         "'" + fName + "'",
                         "does not point to a file"))

    if not isFileOpenable(fName):
        return ' '.join(("The resolved link path",
                         "'" + fName + "'",
                         "does not point to a file which Codimension can open"))
    return None


# Used when the user clicks on a link
def resolveLinkPath(link, fromFile):
    """Resolves the link to the another file"""
    effectiveLink, anchor = splitLinkPath(link)
    fName, tryPaths = resolveFile(effectiveLink, fromFile)

    # fName is not None => file exists on FS
    if not fName:
        logging.error("The link '%s' does not point to an existing file. "
                      "Resolve tries: %s",effectiveLink, ', '.join(tryPaths))
        return None, None

    errMsg = checkExistingLinkTarget(fName)
    if errMsg:
        logging.error(errMsg)
        return None, None

    # All is good: it is a file, it is openable
    return fName, anchor


# Used in a dialog to suggest validity of the further operation
def preResolveLinkPath(link, fromFile, canBeCreated):
    """Tries to resolve the link and repors an error if any"""
    effectiveLink, anchor = splitLinkPath(link)
    fName, tryPaths = resolveFile(effectiveLink, fromFile)

    # Here: the file must not exist
    # If it exists then it should be a file and openable
    # If it does not then it should be creatable
    if fName:
        errMsg = checkExistingLinkTarget(fName)
        if errMsg:
            return None, None, errMsg
        return fName, anchor, None

    if not tryPaths:
        return None, None, \
               "The link '" + effectiveLink + \
               "' is invalid. Use an absolute path or save the file first."

    if canBeCreated:
        for path in tryPaths:
            canCreate, _ = isCreatable(path)
            if canCreate:
                # Can be created here
                return path, anchor, None

        return None, None, \
           "The link '" + effectiveLink + \
           "' points to non existing file which cannot be created due to " \
           "lack of permissions or invalid path. Tried paths: " + \
           ", ".join(tryPaths)

    return None, None, \
        "The link '" + effectiveLink + \
        "' does not point to an existing file"

def printStack():
    """Prints the stack in the log window"""
    for line in traceback.format_stack():
        print(line.strip())


def printReferences(obj):
    """Prints the object references"""
    if obj is None:
        print('No object to print references')
    else:
        refs = gc.get_referrers(obj)
        print('Object references (total ' + str(len(refs)) + '):')
        for item in refs:
            print('---')
            print(type(item))
            val = repr(item)
            if len(val) > 512:
                print('Truncated: ' + val[:512])
            else:
                print(val)




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Control flow UI widget: handling keyboard events"""

from sys import maxsize
from ui.qt import Qt, QGraphicsScene
from flowui.cellelement import CellElement
from flowui.scopeitems import ScopeCellElement
from utils.settings import Settings


CTRL_SHIFT = int(Qt.ShiftModifier | Qt.ControlModifier)
SHIFT = int(Qt.ShiftModifier)
CTRL = int(Qt.ControlModifier)
ALT = int(Qt.AltModifier)
CTRL_KEYPAD = int(Qt.KeypadModifier | Qt.ControlModifier)
NO_MODIFIER = int(Qt.NoModifier)


class CFSceneKeyboardMixin:

    """Encapsulates keyboard handling and related functionality"""

    def __init__(self):
        self.__hotKeys = {
            CTRL: {
                Qt.Key_QuoteLeft: self.highlightInText,
                Qt.Key_Home: self.scrollToTop,
                Qt.Key_End: self.scrollToBottom,
                Qt.Key_A: self.selectAll,
                Qt.Key_Minus: Settings().onFlowZoomOut,
                Qt.Key_Equal: Settings().onFlowZoomIn,
                Qt.Key_0: Settings().onFlowZoomReset},
            NO_MODIFIER: {
                Qt.Key_Home: self.scrollToHBegin,
                Qt.Key_End: self.scrollToHEnd,
                Qt.Key_Escape: self.clearSelection}}

    def keyPressEvent(self, event):
        """Handles the key press event"""
        key = event.key()
        modifiers = int(event.modifiers())
        if modifiers in self.__hotKeys:
            if key in self.__hotKeys[modifiers]:
                self.__hotKeys[modifiers][key]()
                event.accept()
                return
        QGraphicsScene.keyPressEvent(self, event)

    def highlightInText(self):
        """Sync the text with the graphics"""
        firstItem = self.getFirstLogicalItem()
        if firstItem:
            self.clearSelection()
            firstItem.setSelected(True)
            self.parent().view().scrollTo(firstItem)
            firstItem.mouseDoubleClickEvent(None)

    def getFirstLogicalItem(self):
        """Provides the first visible on the screen logical item"""
        view = self.parent().view()
        visibleRect = view.getVisibleRect()

        firstLine = visibleRect.y()

        candidateAfter = None
        candidateAfterDistance = maxsize
        candidateBefore = None
        candidateBeforeDistance = maxsize * -1
        for item in self.items():
            if not isinstance(item, CellElement):
                continue
            if item.isProxyItem():
                continue

            itemY = item.boundingRect().topLeft().y()

            dist = itemY - firstLine
            if dist > 0:
                if dist < candidateAfterDistance:
                    candidateAfterDistance = dist
                    candidateAfter = item
            elif dist > candidateBeforeDistance:
                candidateBeforeDistance = dist
                candidateBefore = item

        item = None
        if candidateAfter:
            item = candidateAfter
        elif candidateBefore:
            item = candidateBefore
        else:
            # No suitable item to select
            return None

        logicalItem = item
        if item.scopedItem():
            if item.subKind in [ScopeCellElement.DECLARATION]:
                logicalItem = item.getTopLeftItem()
        return logicalItem

    def scrollToTop(self):
        """Scrolls the view to the top"""
        view = self.parent().view()
        view.horizontalScrollBar().setValue(0)
        view.verticalScrollBar().setValue(0)

    def scrollToBottom(self):
        """Scrolls the view to the bottom"""
        view = self.parent().view()
        view.horizontalScrollBar().setValue(0)
        view.verticalScrollBar().setValue(view.verticalScrollBar().maximum())

    def scrollToHBegin(self):
        """Scrolls horizontally to the very beginning"""
        view = self.parent().view()
        view.horizontalScrollBar().setValue(0)

    def scrollToHEnd(self):
        """Scrolls horizontally to the very end"""
        view = self.parent().view()
        view.horizontalScrollBar().setValue(
            view.horizontalScrollBar().maximum())

    def selectAll(self):
        """Selects all"""
        moduleItem = None
        for item in self.items():
            if not isinstance(item, CellElement):
                continue
            if item.isProxyItem():
                continue
            if item.kind == CellElement.FILE_SCOPE:
                if item.subKind == ScopeCellElement.TOP_LEFT:
                    moduleItem = item
                    break

        if moduleItem:
            self.clearSelection()
            for item in self.findItemsForRef(moduleItem.ref):
                self.addToSelection(item)

    def onZoomOut(self):
        """Zoom out the view"""
        view = self.parent().view()
        view.zoomOut()

    def onZoomIn(self):
        """Zoom in the view"""
        view = self.parent().view()
        view.zoomIn()
